# -*- coding: utf-8 -*-
"""
The module converts a noncommutative optimization problem provided in SymPy
format to an SDPA semidefinite programming problem.

Created on Sun May 26 15:06:17 2013

@author: Peter Wittek
"""
from __future__ import division, print_function
import sys
from functools import partial
import numpy as np
from scipy.sparse import lil_matrix
from sympy import S, Expr, Add
import time

try:
    from itertools import imap
except ImportError:
    imap = map
try:
    from multiprocessing import Pool, cpu_count
except ImportError:
    pass
try:
    from IPython.core.display import clear_output
    have_ipython = True
except ImportError:
    have_ipython = False

from .nc_utils import apply_substitutions, \
    assemble_monomial_and_do_substitutions, convert_relational, \
    find_variable_set, flatten, flip_sign, get_all_monomials, is_number_type, \
    is_pure_substitution_rule, iscomplex, moment_of_entry, ncdegree, \
    pick_monomials_up_to_degree, save_monomial_index, separate_scalar_factor, \
    simplify_polynomial, unique
from .solver_common import find_solution_ranks, get_sos_decomposition, \
    get_xmat_value, solve_sdp, extract_dual_value
from .cvxpy_utils import convert_to_cvxpy
from .mosek_utils import convert_to_mosek
from .picos_utils import convert_to_picos
from .sdpa_utils import write_to_sdpa, write_to_human_readable
from .chordal_extension import find_variable_cliques


class Relaxation(object):

    def __init__(self):
        """Constructor for the class.
        """
        self.n_vars = 0
        self.F = None
        self.block_struct = []
        self.obj_facvar = 0
        self.constant_term = 0
        # Variables related to the solution
        self.primal = None
        self.dual = None
        self.x_mat = None
        self.y_mat = None
        self.solution_time = None
        self.status = "unsolved"

    def solve(self, solver=None, solverparameters=None):
        """Call a solver on the SDP relaxation. Upon successful solution, it
        returns the primal and dual objective values along with the solution
        matrices. It also sets these values in the `sdpRelaxation` object,
        along with some status information.

        :param sdpRelaxation: The SDP relaxation to be solved.
        :type sdpRelaxation: :class:`ncpol2sdpa.SdpRelaxation`.
        :param solver: The solver to be called, either `None`, "sdpa", "mosek",
                       "cvxpy", "scs", or "cvxopt". The default is `None`,
                       which triggers autodetect.
        :type solver: str.
        :param solverparameters: Parameters to be passed to the solver. Actual
                                 options depend on the solver:

                                 SDPA:

                                   - `"executable"`:
                                     Specify the executable for SDPA. E.g.,
                                     `"executable":"/usr/local/bin/sdpa"`, or
                                     `"executable":"sdpa_gmp"`
                                   - `"paramsfile"`: Specify the parameter file

                                 Mosek:
                                 Refer to the Mosek documentation. All
                                 arguments are passed on.

                                 Cvxopt:
                                 Refer to the PICOS documentation. All
                                 arguments are passed on.

                                 Cvxpy:
                                 Refer to the Cvxpy documentation. All
                                 arguments are passed on.

                                 SCS:
                                 Refer to the Cvxpy documentation. All
                                 arguments are passed on.

        :type solverparameters: dict of str.
        """
        if self.F is None:
            raise Exception("Relaxation is not generated yet. Call "
                            "'SdpRelaxation.get_relaxation' first")
        solve_sdp(self, solver, solverparameters)


class SdpRelaxation(Relaxation):

    """Class for obtaining sparse SDP relaxation.

    :param variables: Commutative or noncommutative, Hermitian or nonhermiatian
                      variables, possibly a list of list of variables if the
                      hierarchy is not NPA.
    :type variables: list of :class:`sympy.physics.quantum.operator.Operator`
                     or
                     :class:`sympy.physics.quantum.operator.HermitianOperator`
                     or a list of list.
    :param parameters: Optional symbolic variables for which moments are not
                       generated.
    :type parameters: list of :class:`sympy.physics.quantum.operator.Operator`
                     or
                     :class:`sympy.physics.quantum.operator.HermitianOperator`
                     or a list of list.
    :param verbose: Optional parameter for level of verbosity:

                       * 0: quiet (default)
                       * 1: verbose
                       * 2: debug level
    :type verbose: int.
    :param normalized: Optional parameter for changing the normalization of
                       states over which the optimization happens. Turn it off
                       if further processing is done on the SDP matrix before
                       solving it.
    :type normalized: bool.
    :param parallel: Optional parameter for allowing parallel computations.
    :type parallel: bool.
    :param number_of_threads: set the number of threads if parallel is True. If None, use cpu_count() of multiprocess.
    :type number_of_threads: int.

    Attributes:
      - `monomial_sets`: The monomial sets that generate the moment matrix blocks.

      - `monomial_index`: Dictionary that maps monomials to SDP variables.

      - `constraints`: The complete set of constraints after preprocesssing.

      - `primal`: The primal optimal value.

      - `dual`: The dual optimal value.

      - `x_mat`: The primal solution matrix.

      - `y_mat`: The dual solution matrix.

      - `solution_time`: The amount of time taken to solve the relaxation.

      - `status`: The solution status of the relaxation.

    """
    def __init__(self, variables, parameters=None, verbose=0, normalized=True,
                 parallel=False, number_of_threads=None):
        """Constructor for the class.
        """
        super(SdpRelaxation, self).__init__()

        self.verbose = verbose

        # Dictionary that maps monomials to SDP variables.
        self.monomial_index = {}

        # Variables related to generating the moment matrix
        self.var_offsets = [0]
        self.variables = []
        self.normalized = normalized
        self.substitutions = {}
        self.pure_substitution_rules = True
        self.monomial_sets = []
        self.level = 0
        self.moment_substitutions = {}
        self.complex_matrix = False

        # Variables related to processing constraints
        self.localizing_monomial_sets = None
        self.constraint_starting_block = 0
        self.constraints = []
        self._constraint_to_block_index = {}
        self._moment_equalities = []
        self._n_inequalities = 0

        # Variables related to basis transformation
        self._original_F = None
        self._original_obj_facvar = 0
        self._original_constant_term = 0
        self._new_basis = None
        self._time0 = None
        n_noncommutative_hermitian = 0
        n_noncommutative_nonhermitian = 0
        n_commutative_hermitian = 0
        n_commutative_nonhermitian = 0
        if isinstance(variables, list):
            if len(variables) > 0 and (isinstance(variables[0], list) or
                                       isinstance(variables[0], tuple)):
                self.variables = [unique(vs) for vs in variables]
            else:
                self.variables = unique(variables)
        else:
            self.variables = [variables]
        for vs in self.variables:
            if not isinstance(vs, list) and not isinstance(vs, tuple):
                vs = [vs]
            for v in vs:
                if v.is_commutative and (v.is_hermitian is None or
                                         v.is_hermitian):
                    n_commutative_hermitian += 1
                elif v.is_commutative:
                    n_commutative_nonhermitian += 1
                elif not v.is_commutative and (v.is_hermitian is None or
                                               v.is_hermitian):
                    n_noncommutative_hermitian += 1
                else:
                    n_noncommutative_nonhermitian += 1
        self.parameters = parameters
        info = ""
        if n_commutative_hermitian > 0:
            info += str(n_commutative_hermitian) + " commuting"
        if n_commutative_nonhermitian > 0:
            if len(info) > 0:
                info += ", "
            info += str(n_commutative_nonhermitian) + \
                " commuting nonhermitian"
        if n_noncommutative_hermitian > 0:
            if len(info) > 0:
                info += ", "
            info += str(n_noncommutative_hermitian) + \
                " noncommuting Hermitian"
        if n_noncommutative_nonhermitian > 0:
            if len(info) > 0:
                info += ", "
            info += str(n_noncommutative_nonhermitian) + \
                " noncommuting nonhermitian"
        if len(info) > 0:
            info += " variables"
        else:
            info += "0 variables"
        info = "The problem has " + info
        if parameters is not None:
            info += ", and " + str(len(flatten(parameters))) + \
                "symbolic parameters"
        if self.verbose > 0:
            print(info)
        self._parallel = False
        if parallel:
            try:
                if number_of_threads is None:
                    self.number_of_threads = cpu_count()
                else:
                    self.number_of_threads = number_of_threads
                self._parallel = parallel
                if self.verbose > 0:
                    print("Parallel processing on %d cores" % self.number_of_threads)
            except:
                print("Warning: multiprocessing cannot be imported!",
                      file=sys.stderr)

    ########################################################################
    # ROUTINES RELATED TO GENERATING THE MOMENT MATRICES                   #
    ########################################################################

    def _process_monomial(self, monomial, n_vars):
        """Process a single monomial when building the moment matrix.
        """
        processed_monomial, coeff = separate_scalar_factor(monomial)
        # Are we substituting this moment?
        try:
            substitute = self.moment_substitutions[processed_monomial]
            if not isinstance(substitute, (int, float, complex)):
                result = []
                if not isinstance(substitute, Add):
                    args = [substitute]
                else:
                    args = substitute.args
                for arg in args:
                    if is_number_type(arg):
                        if iscomplex(arg):
                            result.append((0, coeff*complex(arg)))
                        else:
                            result.append((0, coeff*float(arg)))
                    else:
                        result += [(k, coeff*c2)
                                   for k, c2 in self._process_monomial(arg,
                                                                       n_vars)]
            else:
                result = [(0, coeff*substitute)]
        except KeyError:
            # Have we seen this monomial before?
            try:
                # If yes, then we improve sparsity by reusing the
                # previous variable to denote this entry in the matrix
                k = self.monomial_index[processed_monomial]
            except KeyError:
                # If no, it still may be possible that we have already seen its
                # conjugate. If the problem is real-valued, a monomial and its
                # conjugate should be equal (Hermiticity becomes symmetry)
                if not self.complex_matrix:
                    try:
                    # If we have seen the conjugate before, we just use the
                    # conjugate monomial instead
                        processed_monomial_adjoint = \
                              apply_substitutions(processed_monomial.adjoint(),
                                                  self.substitutions)
                        k = self.monomial_index[processed_monomial_adjoint]
                    except KeyError:
                        # Otherwise we define a new entry in the associated
                        # array recording the monomials, and add an entry in
                        # the moment matrix
                        k = n_vars + 1
                        self.monomial_index[processed_monomial] = k
                else:
                    k = n_vars + 1
                    self.monomial_index[processed_monomial] = k
            result = [(k, coeff)]
        return result

    def _push_monomial(self, monomial, n_vars, row_offset, rowA, columnA, N,
                       rowB, columnB, lenB, prevent_substitutions=False):
        if monomial in self.moment_substitutions:
            return self._push_monomial(self.moment_substitutions[monomial], n_vars, row_offset,
                                       rowA, columnA, N,
                                       rowB, columnB, lenB, prevent_substitutions)
        if is_number_type(monomial):
            if rowA == 0 and columnA == 0 and rowB == 0 and columnB == 0 and \
                    monomial == 1.0:
                if not self.normalized:
                    n_vars += 1
                    self.F[row_offset + rowA * N*lenB + rowB * N +
                           columnA * lenB + columnB, n_vars] = 1
                else:
                    self.F[row_offset + rowA * N*lenB + rowB * N +
                           columnA*lenB + columnB, 0] = float(self.normalized)
            else:
                self.F[row_offset + rowA * N*lenB + rowB * N +
                       columnA * lenB + columnB, 0] = monomial
        elif monomial.is_Add:
            for element in monomial.args:
                n_vars = self._push_monomial(element, n_vars, row_offset,
                                             rowA, columnA, N,
                                             rowB, columnB, lenB, True)
        elif monomial != 0:
            entries = self._process_monomial(monomial, n_vars)
            for entry in entries:
                k, coeff = entry
                # We push the entry to the moment matrix
                self.F[row_offset + rowA * N*lenB + rowB * N +
                       columnA * lenB + columnB, k] = coeff
                if k > n_vars:
                    n_vars = k
        return n_vars

    def _generate_moment_matrix(self, n_vars, block_index, processed_entries,
                                monomialsA, monomialsB, ppt=False):
        """Generate the moment matrix of monomials.

        Arguments:
        n_vars -- current number of variables
        block_index -- current block index in the SDP matrix
        monomials -- |W_d| set of words of length up to the relaxation level
        """
        row_offset = 0
        if block_index > 0:
            for block_size in self.block_struct[0:block_index]:
                row_offset += block_size ** 2
        N = len(monomialsA)*len(monomialsB)
        func = partial(assemble_monomial_and_do_substitutions,
                       monomialsA=monomialsA, monomialsB=monomialsB, ppt=ppt,
                       substitutions=self.substitutions,
                       pure_substitution_rules=self.pure_substitution_rules)
        if self._parallel:
            pool = Pool()
            # This is just a guess and can be optimized
            chunksize = int(max(int(np.sqrt(len(monomialsA) * len(monomialsB) *
                                        len(monomialsA) / 2) / self.number_of_threads),
                                1))
        for rowA in range(len(monomialsA)):
            if self._parallel:
                iter_ = pool.map(func, [(rowA, columnA, rowB, columnB)
                                        for rowB in range(len(monomialsB))
                                        for columnA in range(rowA,
                                                             len(monomialsA))
                                        for columnB in range((rowA == columnA)*rowB,
                                                             len(monomialsB))],
                                 chunksize)
                print_criterion = processed_entries + len(iter_)
            else:
                iter_ = imap(func, [(rowA, columnA, rowB, columnB)
                                    for columnA in range(rowA, len(monomialsA))
                                    for rowB in range(len(monomialsB))
                                    for columnB in range((rowA == columnA)*rowB,
                                                         len(monomialsB))])
            for columnA, rowB, columnB, monomial in iter_:
                processed_entries += 1
                n_vars = self._push_monomial(monomial, n_vars,
                                             row_offset, rowA,
                                             columnA, N, rowB,
                                             columnB, len(monomialsB),
                                             prevent_substitutions=True)
                if self.verbose > 0 and (not self._parallel or
                                         processed_entries == self.n_vars or
                                         processed_entries == print_criterion):
                    percentage = processed_entries / self.n_vars
                    time_used = time.time()-self._time0
                    eta = (1.0 / percentage) * time_used - time_used
                    hours = int(eta/3600)
                    minutes = int((eta-3600*hours)/60)
                    seconds = eta-3600*hours-minutes*60

                    msg = ""
                    if self.verbose > 1 and self._parallel:
                        msg = ", working on block {:0} with {:0} processes with a chunksize of {:0d}"\
                              .format(block_index, self.number_of_threads,
                                      chunksize)
                    msg = "{:0} (done: {:.2%}, ETA {:02d}:{:02d}:{:03.1f}"\
                          .format(n_vars, percentage, hours, minutes, seconds) + \
                          msg
                    msg = "\r\x1b[KCurrent number of SDP variables: " + msg + ")"
                    sys.stdout.write(msg)
                    sys.stdout.flush()

        if self._parallel:
            pool.close()
            pool.join()
        if self.verbose > 0:
            sys.stdout.write("\r")
        return n_vars, block_index + 1, processed_entries

    def _generate_all_moment_matrix_blocks(self, n_vars, block_index):
        processed_entries = 0
        for monomials in self.monomial_sets:
            if len(monomials) > 0 and isinstance(monomials[0], list):
                if len(monomials[0]) != len(monomials[1]):
                    raise Exception("Cannot generate square block from "
                                    "unequal monomial lists!")
                n_vars, block_index, processed_entries = \
                    self._generate_moment_matrix(
                        n_vars,
                        block_index,
                        processed_entries,
                        monomials[0], monomials[1])
            else:
                n_vars, block_index, processed_entries = \
                    self._generate_moment_matrix(
                        n_vars,
                        block_index,
                        processed_entries,
                        monomials, [S.One])
            self.var_offsets.append(n_vars)
        return n_vars, block_index

    ########################################################################
    # ROUTINES RELATED TO GENERATING THE LOCALIZING MATRICES AND PROCESSING#
    # CONSTRAINTS                                                          #
    ########################################################################

    def _get_index_of_monomial(self, element, enablesubstitution=True,
                               daggered=False):
        """Returns the index of a monomial.
        """
        result = []
        processed_element, coeff1 = separate_scalar_factor(element)
        if processed_element in self.moment_substitutions:
            r = self._get_index_of_monomial(self.moment_substitutions[processed_element], enablesubstitution)
            return [(k, coeff*coeff1) for k, coeff in r]
        if enablesubstitution:
            processed_element = \
                apply_substitutions(processed_element, self.substitutions,
                                    self.pure_substitution_rules)
        # Given the monomial, we need its mapping L_y(w) to push it into
        # a corresponding constraint matrix
        if is_number_type(processed_element):
            return [(0, coeff1)]
        elif processed_element.is_Add:
            monomials = processed_element.args
        else:
            monomials = [processed_element]
        for monomial in monomials:
            monomial, coeff2 = separate_scalar_factor(monomial)
            coeff = coeff1*coeff2
            if is_number_type(monomial):
                result.append((0, coeff))
                continue
            k = -1
            if monomial != 0:
                if monomial.as_coeff_Mul()[0] < 0:
                    monomial = -monomial
                    coeff = -1.0 * coeff
            try:
                new_element = self.moment_substitutions[monomial]
                r = self._get_index_of_monomial(self.moment_substitutions[new_element], enablesubstitution)
                result += [(k, coeff*coeff3) for k, coeff3 in r]
            except KeyError:
                try:
                    k = self.monomial_index[monomial]
                    result.append((k, coeff))
                except KeyError:
                    if not daggered:
                        dag_result = self._get_index_of_monomial(monomial.adjoint(),
                                                                 daggered=True)
                        result += [(k, coeff0*coeff) for k, coeff0 in dag_result]
                    else:
                        raise RuntimeError("The requested monomial " +
                                           str(monomial) + " could not be found.")
        return result

    def __push_facvar_sparse(self, polynomial, block_index, row_offset, i, j):
        """Calculate the sparse vector representation of a polynomial
        and pushes it to the F structure.
        """
        width = self.block_struct[block_index - 1]
        # Preprocess the polynomial for uniform handling later
        # DO NOT EXPAND THE POLYNOMIAL HERE!!!!!!!!!!!!!!!!!!!
        # The simplify_polynomial bypasses the problem.
        # Simplifying here will trigger a bug in SymPy related to
        # the powers of daggered variables.
        # polynomial = polynomial.expand()
        if is_number_type(polynomial) or polynomial.is_Mul:
            elements = [polynomial]
        else:
            elements = polynomial.as_coeff_mul()[1][0].as_coeff_add()[1]
        # Identify its constituent monomials
        for element in elements:
            results = self._get_index_of_monomial(element)
            # k identifies the mapped value of a word (monomial) w
            for (k, coeff) in results:
                if k > -1 and coeff != 0:
                    self.F[row_offset + i * width + j, k] += coeff

    def _get_facvar(self, polynomial):
        """Return dense vector representation of a polynomial. This function is
        nearly identical to __push_facvar_sparse, but instead of pushing
        sparse entries to the constraint matrices, it returns a dense
        vector.
        """
        facvar = [0] * (self.n_vars + 1)
        # Preprocess the polynomial for uniform handling later
        if is_number_type(polynomial):
            facvar[0] = polynomial
            return facvar
        polynomial = polynomial.expand()
        if polynomial.is_Mul:
            elements = [polynomial]
        else:
            elements = polynomial.as_coeff_mul()[1][0].as_coeff_add()[1]
        for element in elements:
            results = self._get_index_of_monomial(element)
            for (k, coeff) in results:
                facvar[k] += coeff
        return facvar

    def __process_inequalities(self, block_index):
        """Generate localizing matrices

        Arguments:
        inequalities -- list of inequality constraints
        monomials    -- localizing monomials
        block_index -- the current block index in constraint matrices of the
                       SDP relaxation
        """
        initial_block_index = block_index
        row_offsets = [0]
        for block, block_size in enumerate(self.block_struct):
            row_offsets.append(row_offsets[block] + block_size ** 2)

        if self._parallel:
            pool = Pool()
        for k, ineq in enumerate(self.constraints):
            block_index += 1
            monomials = self.localizing_monomial_sets[block_index -
                                                      initial_block_index-1]
            lm = len(monomials)
            if isinstance(ineq, str):
                self.__parse_expression(ineq, row_offsets[block_index-1])
                continue
            if ineq.is_Relational:
                ineq = convert_relational(ineq)
            func = partial(moment_of_entry, monomials=monomials, ineq=ineq,
                           substitutions=self.substitutions)
            if self._parallel and lm > 1:
                chunksize = max(int(np.sqrt(lm*lm/2) /
                                    self.number_of_threads), 1)
                iter_ = pool.map(func, ([row, column] for row in range(lm)
                                        for column in range(row, lm)),
                                 chunksize)
            else:
                iter_ = imap(func, ([row, column] for row in range(lm)
                                    for column in range(row, lm)))
            if block_index > self.constraint_starting_block + \
                    self._n_inequalities and lm > 1:
                is_equality = True
            else:
                is_equality = False
            for row, column, polynomial in iter_:
                if is_equality:
                    row, column = 0, 0
                self.__push_facvar_sparse(polynomial, block_index,
                                          row_offsets[block_index-1],
                                          row, column)
                if is_equality:
                    block_index += 1
            if is_equality:
                block_index -= 1
            if self.verbose > 0:
                sys.stdout.write("\r\x1b[KProcessing %d/%d constraints..." %
                                 (k+1, len(self.constraints)))
                sys.stdout.flush()
        if self._parallel:
            pool.close()
            pool.join()

        if self.verbose > 0:
            sys.stdout.write("\n")
        return block_index

    def __process_equalities(self, equalities, momentequalities):
        """Generate localizing matrices

        Arguments:
        equalities -- list of equality constraints
        equalities -- list of moment equality constraints
        """
        monomial_sets = []
        n_rows = 0
        le = 0
        if equalities is not None:
            for equality in equalities:
                le += 1
                # Find the order of the localizing matrix
                if equality.is_Relational:
                    equality = convert_relational(equality)
                eq_order = ncdegree(equality)
                if eq_order > 2 * self.level:
                    raise Exception("An equality constraint has degree %d. "
                                    "Choose a higher level of relaxation."
                                    % eq_order)
                localization_order = (2 * self.level - eq_order)//2
                index = find_variable_set(self.variables, equality)
                localizing_monomials = \
                    pick_monomials_up_to_degree(self.monomial_sets[index],
                                                localization_order)
                if len(localizing_monomials) == 0:
                    localizing_monomials = [S.One]
                localizing_monomials = unique(localizing_monomials)
                monomial_sets.append(localizing_monomials)
                n_rows += len(localizing_monomials) * \
                    (len(localizing_monomials) + 1) // 2
        if momentequalities is not None:
            for _ in momentequalities:
                le += 1
                monomial_sets.append([S.One])
                n_rows += 1
        A = np.zeros((n_rows, self.n_vars + 1), dtype=self.F.dtype)
        n_rows = 0
        if self._parallel:
            pool = Pool()
        for i, equality in enumerate(flatten([equalities, momentequalities])):
            func = partial(moment_of_entry, monomials=monomial_sets[i],
                           ineq=equality, substitutions=self.substitutions)
            lm = len(monomial_sets[i])
            if self._parallel and lm > 1:
                chunksize = max(int(np.sqrt(lm*lm/2) /
                                    self.number_of_threads), 1)
                iter_ = pool.map(func, ([row, column] for row in range(lm)
                                        for column in range(row, lm)),
                                 chunksize)
            else:
                iter_ = imap(func, ([row, column] for row in range(lm)
                                    for column in range(row, lm)))
            # Process M_y(gy)(u,w) entries
            for row, column, polynomial in iter_:
                # Calculate the moments of polynomial entries
                if isinstance(polynomial, str):
                    self.__parse_expression(equality, -1, A[n_rows])
                else:
                    A[n_rows] = self._get_facvar(polynomial)
                n_rows += 1
                if self.verbose > 0:
                    sys.stdout.write("\r\x1b[KProcessing %d/%d equalities..." %
                                     (i+1, le))
                    sys.stdout.flush()
        if self._parallel:
            pool.close()
            pool.join()

        if self.verbose > 0:
            sys.stdout.write("\n")
        return A

    def __remove_equalities(self, equalities, momentequalities):
        """Attempt to remove equalities by solving the linear equations.
        """
        A = self.__process_equalities(equalities, momentequalities)
        if min(A.shape != np.linalg.matrix_rank(A)):
            print("Warning: equality constraints are linearly dependent! "
                  "Results might be incorrect.", file=sys.stderr)
        if A.shape[0] == 0:
            return
        c = np.array(self.obj_facvar)
        if self.verbose > 0:
            print("QR decomposition...")
        Q, R = np.linalg.qr(A[:, 1:].T, mode='complete')
        n = np.max(np.nonzero(np.sum(np.abs(R), axis=1) > 0)) + 1
        x = np.dot(Q[:, :n], np.linalg.solve(np.transpose(R[:n, :]), -A[:, 0]))
        self._new_basis = lil_matrix(Q[:, n:])
        # Transforming the objective function
        self._original_obj_facvar = self.obj_facvar
        self._original_constant_term = self.constant_term
        self.obj_facvar = self._new_basis.T.dot(c)
        self.constant_term += c.dot(x)
        x = np.append(1, x)
        # Transforming the moment matrix and localizing matrices
        new_F = lil_matrix((self.F.shape[0], self._new_basis.shape[1] + 1))
        new_F[:, 0] = self.F[:, :self.n_vars+1].dot(x).reshape((new_F.shape[0],
                                                                1))
        new_F[:, 1:] = self.F[:, 1:self.n_vars+1].\
            dot(self._new_basis)
        self._original_F = self.F
        self.F = new_F
        self.n_vars = self._new_basis.shape[1]
        if self.verbose > 0:
            print("Number of variables after solving the linear equations: %d"
                  % self.n_vars)

    def __duplicate_momentmatrix(self, original_n_vars, n_vars, block_index):
        self.var_offsets.append(n_vars)
        row_offset = 0
        for block_size in self.block_struct[0:block_index]:
            row_offset += block_size ** 2
        width = self.block_struct[0]
        for row in range(width**2):
            self.F[row_offset + row, n_vars+1:n_vars + original_n_vars+2] =\
                self.F[row, :original_n_vars+1]
        return n_vars + original_n_vars + 1, block_index + 1

    def __add_new_momentmatrix(self, n_vars, block_index):
        self.var_offsets.append(n_vars)
        row_offset = 0
        for block_size in self.block_struct[0:block_index]:
            row_offset += block_size ** 2
        width = self.block_struct[0]
        for i in range(width):
            for j in range(i, width):
                n_vars += 1
                self.F[row_offset + i * width + j, n_vars] = 1
        return n_vars, block_index + 1

    def __impose_ppt(self, block_index):
        row_offset = 0
        for block_size in self.block_struct[0:block_index-1]:
            row_offset += block_size ** 2
        lenA = len(self.monomial_sets[0])
        lenB = len(self.monomial_sets[1])
        N = lenA*lenB
        for rowA in range(lenA):
            for columnA in range(rowA, lenA):
                for rowB in range(lenB):
                    start_columnB = 0
                    if rowA == columnA:
                        start_columnB = rowB
                    for columnB in range(start_columnB, rowB):
                        orig_row = self.F[row_offset + rowA*N*lenB + rowB * N +
                                          columnA * lenB + columnB]
                        self.F[row_offset + rowA*N*lenB + rowB * N +
                               columnA * lenB + columnB] = \
                            self.F[row_offset + rowA*N*lenB + columnB * N +
                                   columnA * lenB + rowB]
                        self.F[row_offset + rowA*N*lenB + columnB * N +
                               columnA * lenB + rowB] = orig_row

    def __add_extra_momentmatrices(self, extramomentmatrices, n_vars,
                                   block_index):
        original_n_vars = n_vars
        if extramomentmatrices is not None:
            for parameters in extramomentmatrices:
                copy = False
                ppt = False
                for parameter in parameters:
                    if parameter == "copy":
                        copy = True
                    if parameter == "ppt":
                        ppt = True
                if copy:
                    n_vars, block_index = \
                        self.__duplicate_momentmatrix(original_n_vars, n_vars,
                                                      block_index)
                else:
                    n_vars, block_index = \
                        self.__add_new_momentmatrix(n_vars, block_index)
                if ppt:
                    self.__impose_ppt(block_index)
        return n_vars, block_index

    def __parse_expression(self, expr, row_offset, line=None):
        if expr.find("]") > -1:
            sexprs = [sexpr for sexpr in expr.split(']') if len(sexpr)>0]
            for sexpr in sexprs:
                startindex = 0
                asterix_position = -1
                if sexpr.startswith('-') or sexpr.startswith('+'):
                    asterix_position = sexpr.find('*')
                    if asterix_position > -1:
                        startindex = asterix_position + 1
                    else:
                        startindex = 1
                ind = sexpr.find('[')
                if ind > -1:
                    idx = sexpr[ind+1:].split(",")
                    i, j = int(idx[0]), int(idx[1])
                    mm_ind = int(sexpr[startindex:ind])
                    if asterix_position > -1:
                        value = float(sexpr[:asterix_position])
                    elif sexpr.startswith('-'):
                        value = -1.0
                    else:
                        value = 1.0
                    base_row_offset = sum([bs**2 for bs in
                                           self.block_struct[:mm_ind]])
                    width = self.block_struct[mm_ind]
                    if row_offset > -1:
                        self.F[row_offset] += \
                            value*self.F[base_row_offset + i*width + j]
                    else:
                        line += value*self.F[base_row_offset + i*width + j,
                                             :self.n_vars+1].toarray()[0]
                else:
                    value = float(sexpr)
                    if row_offset > -1:
                        self.F[row_offset, 0] += value
                    else:
                        line[0] += value

    ########################################################################
    # ROUTINES RELATED TO INITIALIZING DATA STRUCTURES                     #
    ########################################################################

    def _calculate_block_structure(self, inequalities, equalities,
                                   momentinequalities, momentequalities,
                                   extramomentmatrix, removeequalities,
                                   block_struct=None):
        """Calculates the block_struct array for the output file.
        """
        if block_struct is None:
            if self.verbose > 0:
                print("Calculating block structure...")
            self.block_struct = []
            if self.parameters is not None:
                self.block_struct += [1 for _ in self.parameters]
            for monomials in self.monomial_sets:
                if len(monomials) > 0 and isinstance(monomials[0], list):
                    self.block_struct.append(len(monomials[0]))
                else:
                    self.block_struct.append(len(monomials))
            if extramomentmatrix is not None:
                for _ in extramomentmatrix:
                    for monomials in self.monomial_sets:
                        if len(monomials) > 0 and \
                                isinstance(monomials[0], list):
                            self.block_struct.append(len(monomials[0]))
                        else:
                            self.block_struct.append(len(monomials))
        else:
            self.block_struct = block_struct
        degree_warning = False
        if inequalities is not None:
            self._n_inequalities = len(inequalities)
            n_tmp_inequalities = len(inequalities)
        else:
            self._n_inequalities = 0
            n_tmp_inequalities = 0
        constraints = flatten([inequalities])
        if momentinequalities is not None:
            self._n_inequalities += len(momentinequalities)
            constraints += momentinequalities
        if not removeequalities:
            constraints += flatten([equalities])
        monomial_sets = []
        for k, constraint in enumerate(constraints):
            # Find the order of the localizing matrix
            if k < n_tmp_inequalities or k >= self._n_inequalities:
                if isinstance(constraint, str):
                    ineq_order = 2 * self.level
                else:
                    if constraint.is_Relational:
                        constraint = convert_relational(constraint)
                    ineq_order = ncdegree(constraint)
                    if iscomplex(constraint):
                        self.complex_matrix = True
                if ineq_order > 2 * self.level:
                    degree_warning = True
                localization_order = (2*self.level - ineq_order)//2
                if self.level == -1:
                    localization_order = 0
                if self.localizing_monomial_sets is not None and \
                        self.localizing_monomial_sets[k] is not None:
                    localizing_monomials = self.localizing_monomial_sets[k]
                else:
                    index = find_variable_set(self.variables, constraint)
                    localizing_monomials = \
                        pick_monomials_up_to_degree(self.monomial_sets[index],
                                                    localization_order)
                ln = len(localizing_monomials)
                if ln == 0:
                    localizing_monomials = [S.One]
            else:
                localizing_monomials = [S.One]
                ln = 1
            localizing_monomials = unique(localizing_monomials)
            monomial_sets.append(localizing_monomials)
            if k < self._n_inequalities:
                self.block_struct.append(ln)
            else:
                monomial_sets += [None for _ in range(ln*(ln+1)//2-1)]
                monomial_sets.append(localizing_monomials)
                monomial_sets += [None for _ in range(ln*(ln+1)//2-1)]
                self.block_struct += [1 for _ in range(ln*(ln+1))]

        if degree_warning and self.verbose > 0:
            print("A constraint has degree %d. Either choose a higher level "
                  "relaxation or ensure that a mixed-order relaxation has the"
                  " necessary monomials" % (ineq_order), file=sys.stderr)

        if momentequalities is not None:
            for moment_eq in momentequalities:
                self._moment_equalities.append(moment_eq)
                if not removeequalities:
                    monomial_sets += [[S.One], [S.One]]
                    self.block_struct += [1, 1]
        self.localizing_monomial_sets = monomial_sets

    def __generate_monomial_sets(self, extramonomials):
        if self.level == -1:
            if extramonomials is None or extramonomials == []:
                raise Exception("Cannot build relaxation at level -1 without \
                                monomials specified.")
            if isinstance(extramonomials[0], list):
                self.monomial_sets = extramonomials
            else:
                self.monomial_sets.append(extramonomials)
            return
        if isinstance(self.variables[0], list):
            k = 0
            for variables in self.variables:
                extramonomials_ = None
                if extramonomials is not None:
                    extramonomials_ = extramonomials[k]
                self.monomial_sets.append(
                    get_all_monomials(variables, extramonomials_,
                                      self.substitutions, self.level))
                k += 1
        else:
            if extramonomials is not None and len(extramonomials) > 0 and \
                    isinstance(extramonomials[0], list):
                self.monomial_sets.append(
                    get_all_monomials(self.variables, extramonomials[0],
                                      self.substitutions, self.level))
                self.monomial_sets += extramonomials[1:]
            else:
                self.monomial_sets.append(
                    get_all_monomials(self.variables, extramonomials,
                                      self.substitutions, self.level))

    def _estimate_n_vars(self):
        self.n_vars = 0
        if self.parameters is not None:
            self.n_vars = len(self.parameters)
        for monomials in self.monomial_sets:
            if len(monomials) > 0 and isinstance(monomials[0], list):
                n_monomials = len(monomials[0])
            else:
                n_monomials = len(monomials)

            # The minus one compensates for the constant term in the
            # top left corner of the moment matrix
            self.n_vars += int(n_monomials * (n_monomials + 1) / 2)
            if self.normalized:
                self.n_vars -= 1

    def __add_parameters(self):
        new_n_vars, block_index = 0, 0
        if self.parameters is not None:
            block_index = 1
            for var in self.parameters:
                new_n_vars += 1
                self.monomial_index[var] = new_n_vars
                self.F[new_n_vars - 1, new_n_vars] = 1
        return new_n_vars, block_index

    def __wipe_F_from_constraints(self):
        row_offset = 0
        for block in range(self.constraint_starting_block):
            row_offset += self.block_struct[block]**2
        for row in range(row_offset, len(self.F.rows)):
            self.F.rows[row] = []
            self.F.data[row] = []

    ########################################################################
    # PUBLIC ROUTINES EXPOSED TO THE USER                                  #
    ########################################################################

    def process_constraints(self, inequalities=None, equalities=None,
                            momentinequalities=None, momentequalities=None,
                            block_index=0, removeequalities=False):
        """Process the constraints and generate localizing matrices. Useful
        only if the moment matrix already exists. Call it if you want to
        replace your constraints. The number of the respective types of
        constraints and the maximum degree of each constraint must remain the
        same.

        :param inequalities: Optional parameter to list inequality constraints.
        :type inequalities: list of :class:`sympy.core.exp.Expr`.
        :param equalities: Optional parameter to list equality constraints.
        :type equalities: list of :class:`sympy.core.exp.Expr`.
        :param momentinequalities: Optional parameter of inequalities defined
                                   on moments.
        :type momentinequalities: list of :class:`sympy.core.exp.Expr`.
        :param momentequalities: Optional parameter of equalities defined
                                 on moments.
        :type momentequalities: list of :class:`sympy.core.exp.Expr`.
        :param removeequalities: Optional parameter to attempt removing the
                                 equalities by solving the linear equations.

        :param removeequalities: Optional parameter to attempt removing the
                                 equalities by solving the linear equations.
        :type removeequalities: bool.
        """
        self.status = "unsolved"
        if block_index == 0:
            if self._original_F is not None:
                self.F = self._original_F
                self.obj_facvar = self._original_obj_facvar
                self.constant_term = self._original_constant_term
                self.n_vars = len(self.obj_facvar)
                self._new_basis = None
            block_index = self.constraint_starting_block
            self.__wipe_F_from_constraints()
        self.constraints = flatten([inequalities])
        self._constraint_to_block_index = {}
        for constraint in self.constraints:
            self._constraint_to_block_index[constraint] = (block_index, )
            block_index += 1
        if momentinequalities is not None:
            for mineq in momentinequalities:
                self.constraints.append(mineq)
                self._constraint_to_block_index[mineq] = (block_index, )
                block_index += 1
        if not (removeequalities or equalities is None):
            # Equalities are converted to pairs of inequalities
            for k, equality in enumerate(equalities):
                if equality.is_Relational:
                    equality = convert_relational(equality)
                self.constraints.append(equality)
                self.constraints.append(-equality)
                ln = len(self.localizing_monomial_sets[block_index-
                                                       self.constraint_starting_block])
                self._constraint_to_block_index[equality] = (block_index,
                                                             block_index+ln*(ln+1)//2)
                block_index += ln*(ln+1)
        if momentequalities is not None and not removeequalities:
            for meq in momentequalities:
                self.constraints += [meq, flip_sign(meq)]
                self._constraint_to_block_index[meq] = (block_index,
                                                        block_index+1)
                block_index += 2
        block_index = self.constraint_starting_block
        self.__process_inequalities(block_index)
        if removeequalities:
            self.__remove_equalities(equalities, momentequalities)

    def set_objective(self, objective, extraobjexpr=None):
        """Set or change the objective function of the polynomial optimization
        problem.

        :param objective: Describes the objective function.
        :type objective: :class:`sympy.core.expr.Expr`
        :param extraobjexpr: Optional parameter of a string expression of a
                             linear combination of moment matrix elements to be
                             included in the objective function
        :type extraobjexpr: str.
        """
        if objective is not None:
            facvar = \
                self._get_facvar(simplify_polynomial(objective,
                                                     self.substitutions))
            self.obj_facvar = facvar[1:]
            self.constant_term = facvar[0]
            if self.verbose > 0 and facvar[0] != 0:
                print("Warning: The objective function has a non-zero %s "
                      "constant term. It is not included in the SDP objective."
                      % facvar[0], file=sys.stderr)
        else:
            self.obj_facvar = self._get_facvar(0)[1:]
        if extraobjexpr is not None:
            for sub_expr in extraobjexpr.split(']'):
                startindex = 0
                asterix_position = -1
                if sub_expr.startswith('-') or sub_expr.startswith('+'):
                    asterix_position = sub_expr.find('*')
                    if asterix_position > -1:
                        startindex = asterix_position + 1
                    else:
                        startindex = 1
                ind = sub_expr.find('[')
                if ind > -1:
                    idx = sub_expr[ind+1:].split(",")
                    i, j = int(idx[0]), int(idx[1])
                    mm_ind = int(sub_expr[startindex:ind])
                    if asterix_position > -1:
                        value = float(sub_expr[:asterix_position])
                    elif sub_expr.startswith('-'):
                        value = -1.0
                    else:
                        value = 1.0
                    base_row_offset = sum([bs**2 for bs in
                                           self.block_struct[:mm_ind]])
                    width = self.block_struct[mm_ind]
                    for column in self.F[base_row_offset + i*width + j].rows[0]:
                        self.obj_facvar[column-1] = \
                            value*self.F[base_row_offset + i*width + j, column]

    def __getitem__(self, index):
        """Obtained the value for a polynomial in a solved relaxation.

        :param index: The polynomial.
        :type index: `sympy.core.exp.Expr`

        :returns: The value of the polynomial extracted from the solved SDP.
        :rtype: float
        """
        if not isinstance(index, Expr):
            raise Exception("Not a monomial or polynomial!")
        elif self.status == "unsolved":
            raise Exception("SDP relaxation is not solved yet!")
        else:
            return get_xmat_value(index, self)

    def get_sos_decomposition(self, threshold=0.0):
        """Given a solution of the dual problem, it returns the SOS
        decomposition.

        :param threshold: Optional parameter for specifying the threshold value
                          below which the eigenvalues and entries of the
                          eigenvectors are disregarded.
        :type threshold: float.
        :returns: The SOS decomposition of [sigma_0, sigma_1, ..., sigma_m]
        :rtype: list of :class:`sympy.core.exp.Expr`.
        """
        return get_sos_decomposition(self, threshold=threshold)

    def extract_dual_value(self, monomial, blocks=None):
        """Given a solution of the dual problem and a monomial, it returns the
        inner product of the corresponding coefficient matrix and the dual
        solution. It can be restricted to certain blocks.

        :param monomial: The monomial for which the value is requested.
        :type monomial: :class:`sympy.core.exp.Expr`.
        :param monomial: The monomial for which the value is requested.
        :type monomial: :class:`sympy.core.exp.Expr`.
        :param blocks: Optional parameter to specify the blocks to be included.
        :type blocks: list of `int`.
        :returns: The value of the monomial in the solved relaxation.
        :rtype: float.
        """
        return extract_dual_value(self, monomial, blocks)

    def find_solution_ranks(self, xmat=None, baselevel=0):
        """Helper function to detect rank loop in the solution matrix.

        :param sdpRelaxation: The SDP relaxation.
        :type sdpRelaxation: :class:`ncpol2sdpa.SdpRelaxation`.
        :param x_mat: Optional parameter providing the primal solution of the
                      moment matrix. If not provided, the solution is extracted
                      from the sdpRelaxation object.
        :type x_mat: :class:`numpy.array`.
        :param base_level: Optional parameter for specifying the lower level
                           relaxation for which the rank loop should be tested
                           against.
        :type base_level: int.
        :returns: list of int -- the ranks of the solution matrix with in the
                  order of increasing degree.
        """
        return find_solution_ranks(self, xmat=xmat, baselevel=baselevel)

    def get_dual(self, constraint, ymat=None):
        """Given a solution of the dual problem and a constraint of any type,
        it returns the corresponding block in the dual solution. If it is an
        equality constraint that was converted to a pair of inequalities, it
        returns a two-tuple of the matching dual blocks.

        :param constraint: The constraint.
        :type index: `sympy.core.exp.Expr`
        :param y_mat: Optional parameter providing the dual solution of the
                      SDP. If not provided, the solution is extracted
                      from the sdpRelaxation object.
        :type y_mat: :class:`numpy.array`.
        :returns: The corresponding block in the dual solution.
        :rtype: :class:`numpy.array` or a tuple thereof.
        """
        if not isinstance(constraint, Expr):
            raise Exception("Not a monomial or polynomial!")
        elif self.status == "unsolved" and ymat is None:
            raise Exception("SDP relaxation is not solved yet!")
        elif ymat is None:
            ymat = self.y_mat
        index = self._constraint_to_block_index.get(constraint)
        if index is None:
            raise Exception("Constraint is not in the dual!")
        if len(index) == 2:
            return ymat[index[0]], self.y_mat[index[1]]
        else:
            return ymat[index[0]]

    def write_to_file(self, filename, filetype=None):
        """Write the relaxation to a file.

        :param filename: The name of the file to write to. The type can be
                         autodetected from the extension: .dat-s for SDPA,
                         .task for mosek or .csv for human readable format.
        :type filename: str.
        :param filetype: Optional parameter to define the filetype. It can be
                         "sdpa" for SDPA , "mosek" for Mosek, or "csv" for
                         human readable format.
        :type filetype: str.
        """
        if filetype == "sdpa" and not filename.endswith(".dat-s"):
            raise Exception("SDPA files must have .dat-s extension!")
        if filetype == "mosek" and not filename.endswith(".task"):
            raise Exception("Mosek files must have .task extension!")
        elif filetype is None and filename.endswith(".dat-s"):
            filetype = "sdpa"
        elif filetype is None and filename.endswith(".csv"):
            filetype = "csv"
        elif filetype is None and filename.endswith(".task"):
            filetype = "mosek"
        elif filetype is None:
            raise Exception("Cannot detect filetype from extension!")

        if filetype == "sdpa":
            write_to_sdpa(self, filename)
        elif filetype == "mosek":
            task = convert_to_mosek(self)
            task.writedata(filename)
        elif filetype == "csv":
            write_to_human_readable(self, filename)
        else:
            raise Exception("Unknown filetype")

    def save_monomial_index(self, filename):
        """Write the monomial index to a file.

        :param filename: The name of the file to write to.
        :type filename: str.
        """
        save_monomial_index(filename, self.monomial_index)

    def convert_to_cvxpy(self):
        """Convert an SDP relaxation to a CVXPY problem.

        :returns: :class:`cvxpy.Problem`.
        """
        return convert_to_cvxpy(self)

    def convert_to_picos(self, duplicate_moment_matrix=False):
        """Convert the SDP relaxation to a PICOS problem such that the exported
        .dat-s file is extremely sparse, there is not penalty imposed in terms
        of SDP variables or number of constraints. This conversion can be used
        for imposing extra constraints on the moment matrix, such as partial
        transpose.

        :param duplicate_moment_matrix: Optional parameter to add an
                                        unconstrained moment matrix to the
                                        problem with the same structure as the
                                        moment matrix with the PSD constraint.
        :type duplicate_moment_matrix: bool.

        :returns: :class:`picos.Problem`.
        """
        return \
            convert_to_picos(self,
                             duplicate_moment_matrix=duplicate_moment_matrix)

    def convert_to_mosek(self):
        """Convert an SDP relaxation to a MOSEK task.

        :returns: :class:`mosek.Task`.
        """
        return convert_to_mosek(self)

    def get_relaxation(self, level, objective=None, inequalities=None,
                       equalities=None, substitutions=None,
                       momentinequalities=None, momentequalities=None,
                       momentsubstitutions=None,
                       removeequalities=False, extramonomials=None,
                       extramomentmatrices=None, extraobjexpr=None,
                       localizing_monomials=None, chordal_extension=False):
        """Get the SDP relaxation of a noncommutative polynomial optimization
        problem.

        :param level: The level of the relaxation. The value -1 will skip
                      automatic monomial generation and use only the monomials
                      supplied by the option `extramonomials`.
        :type level: int.
        :param obj: Optional parameter to describe the objective function.
        :type obj: :class:`sympy.core.exp.Expr`.
        :param inequalities: Optional parameter to list inequality constraints.
        :type inequalities: list of :class:`sympy.core.exp.Expr`.
        :param equalities: Optional parameter to list equality constraints.
        :type equalities: list of :class:`sympy.core.exp.Expr`.
        :param substitutions: Optional parameter containing monomials that can
                              be replaced (e.g., idempotent variables).
        :type substitutions: dict of :class:`sympy.core.exp.Expr`.
        :param momentinequalities: Optional parameter of inequalities defined
                                   on moments.
        :type momentinequalities: list of :class:`sympy.core.exp.Expr`.
        :param momentequalities: Optional parameter of equalities defined
                                 on moments.
        :type momentequalities: list of :class:`sympy.core.exp.Expr`.
        :param momentsubstitutions: Optional parameter containing moments that
                                    can be replaced.
        :type momentsubstitutions: dict of :class:`sympy.core.exp.Expr`.
        :param removeequalities: Optional parameter to attempt removing the
                                 equalities by solving the linear equations.
        :type removeequalities: bool.
        :param extramonomials: Optional paramter of monomials to be included,
                               on top of the requested level of relaxation.
        :type extramonomials: list of :class:`sympy.core.exp.Expr`.
        :param extramomentmatrices: Optional paramter of duplicating or adding
                               moment matrices.  A new moment matrix can be
                               unconstrained (""), a copy  of the first one
                               ("copy"), and satisfying a partial positivity
                               constraint ("ppt"). Each new moment matrix is
                               requested as a list of string of these options.
                               For instance, adding a single new moment matrix
                               as a copy of the first would be
                               ``extramomentmatrices=[["copy"]]``.
        :type extramomentmatrices: list of list of str.
        :param extraobjexpr: Optional parameter of a string expression of a
                             linear combination of moment matrix elements to be
                             included in the objective function.
        :type extraobjexpr: str.
        :param localizing_monomials: Optional parameter to specify sets of
                                     localizing monomials for each constraint.
                                     The internal order of constraints is
                                     inequalities first, followed by the
                                     equalities. If the parameter is specified,
                                     but for a certain constraint the automatic
                                     localization is requested, leave None in
                                     its place in this parameter.
        :type localizing_monomials: list of list of `sympy.core.exp.Expr`.
        :param chordal_extension: Optional parameter to request a sparse
                                  chordal extension.
        :type chordal_extension: bool.

        """
        if self.level < -1:
            raise Exception("Invalid level of relaxation")
        self.level = level
        if substitutions is None:
            self.substitutions = {}
        else:
            self.substitutions = substitutions
            for lhs, rhs in substitutions.items():
                if not is_pure_substitution_rule(lhs, rhs):
                    self.pure_substitution_rules = False
                if iscomplex(lhs) or iscomplex(rhs):
                    self.complex_matrix = True
        if momentsubstitutions is not None:
            self.moment_substitutions = momentsubstitutions.copy()
            # If we have a real-valued problem, the moment matrix is symmetric
            # and moment substitutions also apply to the conjugate monomials
            if not self.complex_matrix:
                for key, val in self.moment_substitutions.copy().items():
                    adjoint_monomial = apply_substitutions(key.adjoint(),
                                                           self.substitutions)
                    self.moment_substitutions[adjoint_monomial] = val
        if chordal_extension:
            self.variables = find_variable_cliques(self.variables, objective,
                                                   inequalities, equalities,
                                                   momentinequalities,
                                                   momentequalities)
        self.__generate_monomial_sets(extramonomials)
        self.localizing_monomial_sets = localizing_monomials

        # Figure out basic structure of the SDP
        self._calculate_block_structure(inequalities, equalities,
                                        momentinequalities, momentequalities,
                                        extramomentmatrices,
                                        removeequalities)
        self._estimate_n_vars()
        if extramomentmatrices is not None:
            for parameters in extramomentmatrices:
                copy = False
                for parameter in parameters:
                    if parameter == "copy":
                        copy = True
                if copy:
                    self.n_vars += self.n_vars + 1
                else:
                    self.n_vars += (self.block_struct[0]**2)/2
        if self.complex_matrix:
            dtype = np.complex128
        else:
            dtype = np.float64
        self.F = lil_matrix((sum([bs**2 for bs in self.block_struct]),
                                    self.n_vars + 1), dtype=dtype)

        if self.verbose > 0:
            print(('Estimated number of SDP variables: %d' % self.n_vars))
            print('Generating moment matrix...')
        # Generate moment matrices
        new_n_vars, block_index = self.__add_parameters()
        self._time0 = time.time()
        new_n_vars, block_index = \
            self._generate_all_moment_matrix_blocks(new_n_vars, block_index)
        if extramomentmatrices is not None:
            new_n_vars, block_index = \
                self.__add_extra_momentmatrices(extramomentmatrices,
                                                new_n_vars, block_index)
        # The initial estimate for the size of F was overly generous.
        self.n_vars = new_n_vars
        # We don't correct the size of F, because that would trigger
        # memory copies, and extra columns in lil_matrix are free anyway.
        # self.F = self.F[:, 0:self.n_vars + 1]

        if self.verbose > 0:
            print(('Reduced number of SDP variables: %d' % self.n_vars))
        # Objective function
        self.set_objective(objective, extraobjexpr)
        # Process constraints
        self.constraint_starting_block = block_index
        self.process_constraints(inequalities, equalities, momentinequalities,
                                 momentequalities, block_index,
                                 removeequalities)



"""
@authors: Oskar Słowik, Filip Maciejewski
@contact: osslowik@gmail.com
"""

#(PP) this seems to be a messy file, need MO to dig through it
from typing import Dict
import copy
import math
import random

import qutip
import numpy as np
import scipy as sc

from tqdm import tqdm

import pandas as pd
import matplotlib.pyplot as plt

from qrem.common.printer import qprint

#MOVE_TO math.py
def l1_norm(v):
    sum = 0
    for x in v:
        sum += abs(x)
    return sum


#MOVE_TO math.py what about accuracy? what are the types here?
def delta(v, w):
    val = 0
    if v == w:
        val = 1
    return val


#TODO_MO - is it a repetition from any other binary string generator?
def generate_binary_strings(n, l):
    if n == 0:
        return l
    else:
        if len(l) == 0:
            return generate_binary_strings(n - 1, ["0", "1"])
        else:
            return generate_binary_strings(n - 1, [i + "0" for i in l] + [i + "1" for i in l])


#(PP) dobuled in other files
#(PP) part of a model of Bravi noise
#(PP) should go to qrem.noise_modelling
def return_1local_gen(type, r):
    if type == '0':
        gen = [[-r, 0], [r, 0]]
    elif type == '1':
        gen = [[0, r], [0, -r]]
    else:
        raise ValueError('Wrong generator type!')
    return np.array(gen)


#(PP) dobuled in other files
#(PP) part of a model of Bravi noise
#(PP) should go to qrem.noise_modelling
def return_2local_gen(type, r):
    if type == '00':
        gen = [[-r, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [r, 0, 0, 0]]
    elif type == '01':
        gen = [[0, 0, 0, 0], [0, -r, 0, 0], [0, r, 0, 0], [0, 0, 0, 0]]
    elif type == '10':
        gen = [[0, 0, 0, 0], [0, 0, r, 0], [0, 0, -r, 0], [0, 0, 0, 0]]
    elif type == '11':
        gen = [[0, 0, 0, r], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, -r]]
    else:
        raise ValueError('Wrong generator type!')
    return np.array(gen)


'''
s - a list of numpy vectors representing measurement results
O - a numpy vector with values O(x)
S - a list of numpy arrays representing operatrs S_alpha
c - a list of weights of each S_alpha
kronecker_delta - kronecker_delta parameter
'''


def bravyi_algorithm_1_naive(s, O, S, c, delta):
    c_l1_norm = l1_norm(c)
    # print(c_l1_norm)
    M = len(s)  # number of experiments
    # print(M)
    d = O.shape[0]  # dimension
    # print(d)
    T = math.ceil(4 * c_l1_norm ** 2 / (delta ** 2))
    # print(T)
    xi = np.zeros(T)
    q_alpha = [abs(x) / c_l1_norm for x in c]
    # print(q_alpha)
    for t in range(T):
        i = random.randrange(0, M)
        # print("i"+str(i))
        alpha = np.random.choice(np.arange(0, len(c)), p=q_alpha)
        # print(alpha)
        q_x = S[alpha].dot(s[i])
        # print(q_x)
        x = np.random.choice(np.arange(0, d), p=q_x)
        # print(x)
        xi[t] = np.sign(c[alpha]) * O[x]
        # print(xi[t])
    return c_l1_norm * np.sum(xi) / T


'''
O=np.array([0.2,0.3,-1,1])
S=[np.array([[0,0,0,0],[1,0,0,0],[0,0,0,1],[0,1,1,0]]),np.array([[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,1]])]
s=[np.array([0,0,1,0]),np.array([0,0,1,0]),np.array([0,0,1,0]),np.array([0,0,1,0]),np.array([0,0,0,1])]
c=[0.95,0.05]
kronecker_delta=0.1
print(bravyi_algorithm_1(O,S,s,c,kronecker_delta))
'''


# p(y_0...y_k | x_0 ... x_n), k<=n
def get_cond_prob(x, y, G_dict, lambda_par, printing=False):
    if printing: print('look for ' + str(x) + '->' + str(y))
    val = 0
    rank = len(y)
    support = set(range(rank))
    if printing: print('supp: ' + str(support))
    val += delta(y, x[:rank])
    if val == 1:
        if printing: print('DIAGONAL')
    for op_support_ordered_str in G_dict:
        operator = G_dict[op_support_ordered_str]
        if printing: print(operator)
        op_support_ordered = [int(qubit) for qubit in op_support_ordered_str.split('q')[1:]]
        if printing: print(op_support_ordered)

        op_support_cap_support = set(op_support_ordered).intersection(set(support))
        if not op_support_cap_support == set():
            support_cond = True
            if printing: print('oscs: ' + str(op_support_cap_support))
            support_minus_op_support = set(support).difference(set(op_support_ordered))
            for qubit in support_minus_op_support:
                if y[qubit] != x[qubit]:
                    support_cond = False
                    if printing: print('sc failed for qubit' + str(qubit))
                    break
            if (support_cond):
                if printing: print('sc OK')
                op_support_minus_support = set(op_support_ordered).difference(set(support))
                if printing: print('osms: ' + str(op_support_minus_support))
                y_supp_vals = dict()

                for qubit in op_support_cap_support:
                    y_supp_vals[qubit] = y[qubit]

                no_of_y_summed = len(op_support_minus_support)

                if (no_of_y_summed > 0):
                    y_summed_bitstrings = generate_binary_strings(no_of_y_summed, [])
                    for y_summed_bitstr in y_summed_bitstrings:
                        i = 0
                        if printing: print('checking bitstr' + str(y_summed_bitstr))
                        for qubit in op_support_minus_support:
                            y_supp_vals[qubit] = y_summed_bitstr[i]
                            i += 1
                        y_supp_vals_ordered = "".join(
                            "%s" % y_supp_vals[qubit] for qubit in op_support_ordered)
                        row = int(y_supp_vals_ordered, 2)
                        x_supp_vals_ordered = "".join("%s" % x[qubit] for qubit in op_support_ordered)
                        col = int(x_supp_vals_ordered, 2)
                        if printing: print('added val ' + str(operator[row][col] / lambda_par))
                        val += operator[row][col] / lambda_par
                else:
                    y_supp_vals_ordered = "".join("%s" % y[qubit] for qubit in op_support_ordered)
                    row = int(y_supp_vals_ordered, 2)
                    x_supp_vals_ordered = "".join("%s" % x[qubit] for qubit in op_support_ordered)
                    col = int(x_supp_vals_ordered, 2)
                    if printing: print('no added bitstr')
                    if printing: print('added val ' + str(operator[row][col] / lambda_par))
                    val += operator[row][col] / lambda_par

    return val


G_dict = dict()
r01_00 = 10
r01_11 = 9
r01_01 = 8

r12_00 = 7
r12_11 = 6
r12_01 = 5

# test_x=[[0,0,1],[0,0,0],[1,1,1]]
G_dict['q0q1'] = [[-r01_00, 0, 0, r01_11], [0, -r01_01, 0, 0], [0, r01_01, 0, 0],
                  [r01_00, 0, 0, -r01_11]]
G_dict['q0q2'] = [[-r12_00, 0, 0, r12_11], [0, -r12_01, 0, 0], [0, r12_01, 0, 0],
                  [r12_00, 0, 0, -r12_11]]

N = 3

x_set = generate_binary_strings(N, [])
y_set = generate_binary_strings(N - 1, [])
lambda_par = 10

'''
process_mat=dict()
for x in x_set:
    for y in y_set:
        process_mat[(x,y)]=get_cond_prob(x, y, G_dict, lambda_par)
        print(x+'->'+y+' '+str(process_mat[(x,y)]))


ser = pd.Series(list(process_mat.values()),
                  index=pd.MultiIndex.from_tuples(process_mat.keys()))
df = ser.unstack().fillna(0).transpose()
sns.heatmap(df)
plt.savefig("output2.png")
'''


def sample_from_B_pow_n(n, x, G_dict, lambda_par, printing=False):
    sample = x
    for step in range(n):
        if printing: print('pre ' + sample)
        sample = sample_from_B(sample, G_dict, lambda_par)
        if printing: print('post ' + sample)
    return sample


def sample_from_B(x, G_dict, lambda_par, printing=False):
    y = ''
    no_of_bits = len(x)
    for bit_no in range(no_of_bits):
        y_0 = y + '0'
        y_1 = y + '1'
        if printing: print('testing ' + y_0 + ' and ' + y_1)
        cp_0 = get_cond_prob(x, y_0, G_dict, lambda_par)
        cp_1 = get_cond_prob(x, y_1, G_dict, lambda_par)
        tot = cp_0 + cp_1
        p_0 = cp_0 / tot
        if printing: print('p_0 ' + str(p_0))
        val = random.uniform(0, 1)
        if val < p_0:
            y = y + '0'
            if printing: print('extending ' + y)
        else:
            y = y + '1'
            if printing: print('extending ' + y)
    return y


# sample_from_power_of_B_matrix(10, x_set[0], G_dict, lambda_par)


O = {'000': 0, '001': 1, '010': 1, '011': 2, '100': 0, '101': 2, '110': 2, '111': 3}
s = ['000'] * 50
s2 = ['001'] * 50
s.extend(s2)
delta_par = 0.1

'''
s - a list of strings representing measurement results
O - a dict with keys - strings x and values - O(x)
kronecker_delta - kronecker_delta parameter
'''


def bravyi_algorithm_1(s, O, G_dict, delta_par, lambda_par):
    c_l1_norm = math.exp(2 * lambda_par)
    M = len(s)  # number of experiments
    # print('M '+str(M))

    T = math.ceil(4 * (c_l1_norm ** 2) / (delta_par ** 2))
    # print('T ' +str(T))
    sum_xi = 0
    for t in range(T):
        i = random.randrange(0, M)
        # print("i"+str(i))
        n = np.random.poisson(lambda_par)
        # print('n '+str(n))
        x = sample_from_B_pow_n(n, s[i], G_dict, lambda_par)
        # print('x '+x)

        if x in O.keys():
            if (n % 2) == 0:
                sign = 1.0
            else:
                sign = -1.0
            sum_xi += sign * O[x]
            # print(xi[t])
    return c_l1_norm * sum_xi / T


# print(bravyi_algorithm_1(s, O, G_dict, delta_par, lambda_par))



def sample_from_bravyi_noise_model(y, N, G_dict, lambda_par):
    out = []
    for t in tqdm(range(N)):
        n = np.random.poisson(lambda_par)
        # print('n ' + str(n))
        out.append(sample_from_B_pow_n(n, y, G_dict, lambda_par))
    return out


def bravyi_noise_model_naive_distribution(y, G_dict):
    N = len(y)
    G = convert_G_dict_to_G(G_dict, N)
    # print(G)
    # print(np.sum(G, axis=0))
    A = sc.linalg.expm(G)
    # print(A)
    # print(np.sum(A, axis=0))
    y_index = int(y, 2)
    # print(y_index)
    pd_x = A[:, y_index]
    return pd_x

#Double entry in
def construct_model_from_rates(r_dict):
    G_dict = dict()
    for op_support_ordered_str in r_dict:
        no_of_qubits = int(len(op_support_ordered_str) / 2)
        dim = int(math.pow(2, no_of_qubits))
        gen_total = np.zeros((dim, dim))
        for op_type in r_dict[op_support_ordered_str]:
            r = r_dict[op_support_ordered_str][op_type]
            if len(op_type) == 1:
                gen = return_1local_gen(op_type, r)
            elif len(op_type) == 2:
                gen = return_2local_gen(op_type, r)
            else:
                raise ValueError('Wrong generator dimension in r_dict!')
            gen_total = gen_total + gen
        G_dict[op_support_ordered_str] = gen_total
    return G_dict


r_dict = {'q0q1': {'00': 10, '11': 9, '01': 8}, 'q0q2': {'00': 7, '11': 6, '01': 5}}


# print(sample_from_CTMP_noise_model_conditional_probabilities(s, 10, G_dict, lambda_par))


# def generate_bravyi_G_dict(rates,N):

def embed_1q_operator(N, gate, index):
    if index == 0:
        embed_operator = np.kron(gate, np.eye(int(2 ** (N - 1))))
        return embed_operator
    else:
        first_eye = np.eye(2 ** (index))
        second_eye = np.eye(2 ** (N - index - 1))

        embeded_operator = np.kron(np.kron(first_eye, gate), second_eye)

        return embeded_operator


def embed_operator(N, gate, indices=[0, 1]):
    N_small = int(np.log2(gate.shape[0]))

    if N_small == 1:
        return embed_1q_operator(N, gate, indices[0])

    qutip_object = qutip.Qobj(gate, dims=[[2, 2] for i in range(N_small)])

    return qutip.qip.expand_operator(qutip_object, N, indices).full()


# def extend_matrix_element(gen_supp_indices, entry, G, N, supp1='', supp2=''):
#    if len(supp2) == N:
#        row
#        # add element
#        return

#    if len(supp1) == N:
#        extend_matrix_element(gen_supp_indices, entry, G, N, supp1, supp2 + '0')
#        extend_matrix_element(gen_supp_indices, entry, G, N, supp1, supp2 + '1')
#    else:
#        extend_matrix_element(gen_supp_indices, entry, G, N, supp1 + '0', supp2)
#        extend_matrix_element(gen_supp_indices, entry, G, N, supp1 + '1', supp2)

def generate_random_r_dict(N, bound):
    r_dict = dict()
    for qubit_1 in range(N):
        supp_1_local = 'q' + str(qubit_1)
        r_dict_sub_1 = dict()
        for op_type_1 in ['0', '1']:
            r_dict_sub_1[op_type_1] = bound * np.random.random()
        r_dict[supp_1_local] = r_dict_sub_1
        for qubit_2 in range(qubit_1 + 1, N):
            supp_2_local = supp_1_local + 'q' + str(qubit_2)
            r_dict_sub_2 = dict()
            for op_type_2 in ['00', '01', '10', '11']:
                r_dict_sub_2[op_type_2] = bound * np.random.random()
            r_dict[supp_2_local] = r_dict_sub_2
    return r_dict


def generate_random_G_dict(N, bound):
    r_dict = generate_random_r_dict(N, bound)
    G_dict = construct_model_from_rates(r_dict)

    # TODO OS: this can be done inside above function
    lambda_b = get_r_sum_from_r_dict(r_dict)
    return G_dict, lambda_b


# q1 q2 ...
def convert_G_dict_to_G(G_dict, N):
    dim = int(math.pow(2, N))
    G = np.zeros((dim, dim))
    for gen_supp in G_dict:
        gen_supp_indices = [int(qubit) for qubit in gen_supp.split('q')[1:]]
        operator = np.array(G_dict[gen_supp])
        # print(N)
        # print(operator)
        operator_emb = embed_operator(N, operator, indices=gen_supp_indices)
        # print(np.shape(operator_emb))
        G = G + operator_emb
    return G


def get_r_sum_from_r_dict(r_dict):
    r_sum = 0
    for op_support_ordered_str in r_dict:
        for op_type in r_dict[op_support_ordered_str]:
            r = r_dict[op_support_ordered_str][op_type]
            r_sum = r_sum + r
    return r_sum


def test_bravyi_sampling(y, bound, no_of_samples):
    no_of_qubits = len(y)
    dim = int(math.pow(2, no_of_qubits))
    pd_x_approx = np.zeros(dim)
    G_dict, lambda_par = generate_random_G_dict(no_of_qubits, bound)
    pd_x = bravyi_noise_model_naive_distribution(y, G_dict)
    samples = sample_from_bravyi_noise_model(y, no_of_samples, G_dict, lambda_par)
    for sample in samples:
        index = int(sample, 2)
        pd_x_approx[index] += 1
    pd_x_approx = pd_x_approx / len(samples)

    pointwise_error = np.absolute(pd_x_approx - pd_x)
    TVD = np.max(pointwise_error)
    print(f'TVD: {TVD:.5f}')
    # print(samples)
    # print(pd_x)
    # print(pd_x_approx)

    ind = np.arange(dim)
    # print(ind)
    width = 0.35
    plt.bar(ind, pd_x_approx, width, label='sampling_alg')
    plt.bar(ind + width, pd_x, width, label='true')

    plt.ylabel('p')
    plt.title(
        'Bravyi sampling pd for y=' + y + ', samples= ' + str(no_of_samples) + f', TVD= {TVD:.5f}')

    xt = [s[::-1] for s in generate_binary_strings(no_of_qubits, [])]

    plt.xticks(ind + width / 2, xt)

    plt.legend(loc='best')
    plt.show()


# test_bravyi_sampling('0101',0.05, 10000)

'''
N = 3
r_d = generate_random_r_dict(N, 2)
print(r_d)
r_sum = get_r_sum_from_r_dict(r_d)
G_dict = construct_model_from_rates(r_d)
G = convert_G_dict_to_G(G_dict, N)
print(r_sum)
print(G)
'''
# print(convert_G_dict_to_G(G_dict, 3))
# G_d=construct_model_from_rates(r_dict)
# print(convert_G_dict_to_G(G_d, 3))


def get_CTMP_rates_from_results(results_dictionary_ddot: Dict[str, Dict[str, int]],
                                number_of_qubits: int):
    single_qubits = list(range(number_of_qubits))
    pairs = [(i, j) for i in range(number_of_qubits) for j in range(i + 1, number_of_qubits)]

    local_noise_matrices_CTMP = {pair: np.zeros((4, 4), dtype=float) for pair in pairs}

    for pair in pairs:
        pair_complement = list(set(single_qubits).difference(set(pair)))

        for global_input_state, results_dictionary in results_dictionary_ddot.items():
            marginal_input_state = ''.join([global_input_state[x] for x in pair])
            # print(marginal_input_state)
            input_state_complement = [global_input_state[x] for x in pair_complement]

            for global_output_state, ticks in results_dictionary.items():
                marginal_output_state = ''.join([global_output_state[x] for x in pair])
                output_state_complement = [global_output_state[x] for x in pair_complement]

                if output_state_complement == input_state_complement:
                    # check if this is their convention!
                    local_noise_matrices_CTMP[pair][int(marginal_output_state, 2),
                                                    int(marginal_input_state, 2)] += ticks

        # print(local_noise_matrices_CTMP[pair])
        # raise KeyError
    # normalize to stochastic matrices
    for pair in pairs:
        for k in range(4):
            if sum(local_noise_matrices_CTMP[pair][:, k]) != 0:
                local_noise_matrices_CTMP[pair][:, k] /= sum(local_noise_matrices_CTMP[pair][:, k])

    # Get G matrices
    # TODO: potential bug with logarithm branch
    # print(local_noise_matrices_CTMP[pair])
    G_matrices = {pair: sc.linalg.logm(local_noise_matrices_CTMP[pair]) for pair in pairs}

    # ancillary function
    def _chop_negatives(M):
        (m, n) = M.shape

        chopped_M = copy.deepcopy(M)
        for i in range(m):
            for j in range(n):
                if i != j and M[i, j] < 0:
                    chopped_M[i, j] = 0

        return chopped_M

    # Get G' matrices
    G_prime_matrices = {pair: _chop_negatives(G_matrices[pair]) for pair in pairs}

    rates_dictionary_2q = {f"q{pair[0]}q{pair[1]}": {'00': 0,
                                                     '01': 0,
                                                     '10': 0,
                                                     '11': 0} for pair in pairs}

    for pair in pairs:
        G_prime_matrix_now = G_prime_matrices[pair]
        rates_dictionary_2q[f"q{pair[0]}q{pair[1]}"]['00'] = G_prime_matrix_now[3, 0]
        rates_dictionary_2q[f"q{pair[0]}q{pair[1]}"]['01'] = G_prime_matrix_now[2, 1]
        rates_dictionary_2q[f"q{pair[0]}q{pair[1]}"]['10'] = G_prime_matrix_now[1, 2]
        rates_dictionary_2q[f"q{pair[0]}q{pair[1]}"]['11'] = G_prime_matrix_now[0, 3]

    rates_dictionary_1q = {f"q{q}": {'0': 0, '1': 0} for q in single_qubits}
    for qj in single_qubits:
        r0, r1 = 0, 0

        for q_other in list(set(single_qubits).difference(set([qj]))):
            G_prime_matrix_now = G_prime_matrices[tuple(sorted([qj, q_other]))]

            r0 += G_prime_matrix_now[2, 0] + G_prime_matrix_now[3, 1] \
                # +G_prime_matrix_now[1,0]+G_prime_matrix_now[3,2]

            r1 += G_prime_matrix_now[0, 2] + G_prime_matrix_now[1, 3] \
                # +G_prime_matrix_now[0, 1] + G_prime_matrix_now[2, 3]

        r0 /= 2 * (number_of_qubits - 1)
        r1 /= 2 * (number_of_qubits - 1)

        rates_dictionary_1q[f"q{qj}"]['0'] = r0
        rates_dictionary_1q[f"q{qj}"]['1'] = r1

    rates_dictionary = {**rates_dictionary_1q,
                        **rates_dictionary_2q}
    return rates_dictionary


def check_CTMP_consistency(no_of_qubits, bound, samples_per_input):
    r_dict = generate_random_r_dict(no_of_qubits, bound)
    G_dict = construct_model_from_rates(r_dict)

    lambda_par = get_r_sum_from_r_dict(r_dict)

    output = dict()

    for input in generate_binary_strings(no_of_qubits, []):
        ticks = dict()
        input_results = sample_from_bravyi_noise_model(input, samples_per_input, G_dict, lambda_par)

        for result in input_results:
            if result in ticks.keys():
                ticks[result] = ticks[result] + 1
            else:
                ticks[result] = 1

        output[input] = ticks

    r_dict_reconstr = get_CTMP_rates_from_results(output, no_of_qubits)
    return r_dict, r_dict_reconstr


number_of_samples = 10 ** 5
r_dict, r_dict_reconstr = check_CTMP_consistency(3, 0.05, number_of_samples)
print(r_dict)
print(r_dict_reconstr)

from qrem.functions_qrem import ancillary_functions as anf

print()
qprint("Number of samples:", number_of_samples, 'red')
for qubit_key in r_dict.keys():

    dict_now_true, dict_now_estimated = r_dict[qubit_key], r_dict_reconstr[qubit_key]

    qprint("QUBITS:", qubit_key, 'green')
    for rate_key in dict_now_true.keys():
        qprint("Rate:", rate_key, 'red')
        qprint("(True value, Estimated value, Difference)",
                       f"({np.round(dict_now_true[rate_key], 4)},"
                       f"{np.round(dict_now_estimated[rate_key], 4)}, "
                       f"{np.round(dict_now_true[rate_key] - dict_now_estimated[rate_key], 4)})")

    print()



import numpy as np
from qrem.ctmp.mitigtools.mitig_utils import _mitigate
from qrem.types import CTMPModelData
from typing import Dict, Tuple, List



def mitigate_expected_value(observable: List[List],
                            results_dictionary: Dict[str, Dict[str, int]],
                            noise_model: CTMPModelData,
                            T: int) -> float:
    """Mitigation procedure as described in Bravyi et al. Returns mitigated expectation value of 
    the observable.

    model - the CTMP model instance
    exp_results - dictionary of experimental results (each result = string of 0s and 1s)
    observable - observable O whose expected value is to be mitigated, given as a list of coefficients
    in strings of Z, e.g. [ [(), 1], [(0,1,3), -2]] denotes O = I - 2 * ZZIZ
    T - number of timesteps of the algorithm"""
    result = {}
    for k, v in results_dictionary.items():
        result[k] = _mitigate(noise_model, v, observable, T, "obs")
    return result



def mitigate_marginals(marginals_list: List[Tuple],
                       results_dictionary: Dict[str, Dict[str, int]],
                       noise_model: CTMPModelData,
                       T: int,
                       ensure_proper_probability_distribution: bool=False)->Dict[str, Dict[Tuple[int], np.array]]:
    """Mitigation procedure as described in Bravyi et al. Returns list of mitigated marginals.

    model - the CTMP model instance
    exp_results - dictionary of experimental results 
    marginals - list of marginals to be mitigated e.g. [(0,1), (2,3,4)]
    T - number of timesteps of the algorithm
    proj - boolean variable indicating whether resulting marginals should be projected onto physical ones
    """
    result = {}
    for k, v in results_dictionary.items():
        marginal_results = {}
        for marginal in marginals_list:
            marginal_results[marginal] = _mitigate(noise_model, v, marginal, T, "marginal", ensure_proper_probability_distribution)
        result[k] = marginal_results
    return result



#JT: We have a separate file in functions_qrem where clustering algoithms are defined (functions_noise_model_heuristic file). Is this redundant?

#JT There are two heuristic_clustering_algorithm_base calsses (the other in cluststering_algorithms/clustering_methods/child_classes). Is this redundant

from qrem.noise_model_generation.CN.clustering_algorithms.clustering_methods.functions.functions_noise_model_heuristic import partition_algorithm_v1_cummulative
from qrem.noise_model_generation.CN.clustering_algorithms.clustering_methods.functions.functions_noise_model_heuristic import partition_algorithm_v2_cummulative
from qrem.noise_model_generation.CN.clustering_algorithms.clustering_methods.functions.functions_noise_model_heuristic import partition_algorithm_v3_cummulative
from qrem.noise_model_generation.CN.clustering_algorithms.clustering_methods.functions.functions_noise_model_heuristic import partition_algorithm_v4_cummulative

#JT Here we import a new clustering algorithm. The import is from qrem.functions_qrem, functions_noise_model_heuristic file (see above)
from qrem.noise_model_generation.CN.clustering_algorithms.clustering_methods.functions.functions_noise_model_heuristic import partition_algorithm_v4_cummulative_temp


class HeuristicClusteringAlgorithmBase:

    def __init__(self,
                 version=None,
                 correlations_table=None,
                 alpha=None,
                 C_maxsize=None, N_alg=None, N_rand=None,
                 printing=False, drawing=False, disable_pb=None):

        if disable_pb is None:
            # if correlations_table.shape[0]>=100:
            #     disable_pb = False
            # else:
            disable_pb = True

        self.version = version
        self.correlations_table = correlations_table
        self.alpha = alpha
        self.C_maxsize = C_maxsize
        self.N_alg = N_alg
        self.N_rand = N_rand
        self.printing = printing
        self.drawing = drawing
        self.disable_pb = disable_pb
        self.global_best_parition_sorted = None
        self.global_best_value = None

    def clusterize(self, correlations_table=None, alpha=None):
        if correlations_table is not None:
            self.correlations_table = correlations_table
        elif self.correlations_table is None:
            raise ValueError("Correlations_table not specified.")
        if alpha is not None:
            self.alpha = alpha
        elif self.alpha is None:
            raise ValueError("Alpha not specified.")
        if self.C_maxsize is None:
            raise ValueError("C_maxsize not specified.")
        if self.N_alg is None:
            raise ValueError("N_alg not specified.")

        if self.version == 'v1':
            self.global_best_parition_sorted, self.global_best_value = partition_algorithm_v1_cummulative(
                self.correlations_table,
                self.alpha,
                self.C_maxsize,
                self.N_alg,
                self.printing,
                self.drawing, self.disable_pb)
        elif self.version == 'v2':
            if self.N_rand is None:
                raise ValueError("N_rand not specified.")

            self.global_best_parition_sorted, self.global_best_value = partition_algorithm_v2_cummulative(
                self.correlations_table,
                self.alpha,
                self.C_maxsize,
                self.N_alg,
                self.N_rand,
                self.printing,
                self.drawing, self.disable_pb)

        elif self.version == 'v3':
            if self.N_rand is None:
                raise ValueError("N_rand not specified.")

            self.global_best_parition_sorted, self.global_best_value = partition_algorithm_v3_cummulative(
                self.correlations_table,
                self.alpha,
                self.C_maxsize,
                self.N_alg,
                self.N_rand,
                self.printing,
                self.drawing, self.disable_pb)
        elif self.version == 'v4':

            self.global_best_parition_sorted, self.global_best_value = partition_algorithm_v4_cummulative(
                self.correlations_table,
                self.alpha,
                self.C_maxsize,
                self.N_alg,
                self.printing,
                self.drawing, self.disable_pb)

        #JT: We add this to use probabilistic clustering algorithm

        elif self.version == 'v4_temp':

            self.global_best_parition_sorted, self.global_best_value = partition_algorithm_v4_cummulative_temp(
                self.correlations_table,
                self.alpha,
                self.C_maxsize,
                self.N_alg,
                self.printing,
                self.drawing, self.disable_pb,temp=1.0)

        else:
            raise ValueError("Implemented versions are: v1, v2, v3, v4")

        return self.global_best_parition_sorted, self.global_best_value



"""
@author: Jan Tuziemski, Filip Maciejewski, Joanna Majsak, Oskar Słowik, Marcin Kotowski, Katarzyna Kowalczyk-Murynka, Paweł Przewłocki, Piotr Podziemski, Michał Oszmaniec
@contact: filip.b.maciejewski@gmail.com
"""

from collections import defaultdict
from typing import Dict, Optional, Tuple, List

from tqdm import tqdm
import numpy as np

from qrem.functions_qrem import functions_data_analysis as fda, ancillary_functions as anf, quantum_ancillary_functions as quanf
from qrem.noise_mitigation.base_classes.EnergyEstimatorBase import EnergyEstimatorBase
from qrem.noise_model_generation.CTMP import functions_CTMP as fun_CTMP
from qrem.noise_simulation.CTMP import functions_CTMP_conditional as CTMP_cond, \
    functions_CTMP_simple as CTMP_simp


from qrem.common.printer import qprint

class EnergyCorrectorGlobalCTMP(EnergyEstimatorBase):
    """
    This is class for implementing quasiprobability-like error mitigation based on CTMP model [#TODO FBM: add ref].

    """


    def __init__(self,
                 results_dictionary: Dict[str, Dict[str, int]],
                 hamiltonian_dictionaries: Dict[str, Dict[Tuple[int], float]],
                 CTMP_rates_dictionary: Dict[Tuple[int], Dict[str, float]],
                 noise_strength_CTMP: Optional[float] = None,
                 ) -> None:
        """
        :param results_dictionary: see parent class description
        :param hamiltonian_dictionaries: see parent class description
        :param CTMP_rates_dictionary: dictionary of error rates
        """

        super().__init__(results_dictionary=results_dictionary,
                         hamiltonian_dictionaries=hamiltonian_dictionaries)

        self._CTMP_rates_dictionary = CTMP_rates_dictionary
        self._energies_dictionary_corrected = {}
        self._default_max_number_of_samples = 10**5

        self._number_of_qubits = len(list(list(list(results_dictionary.values())[0].keys())[0]))
        if noise_strength_CTMP is None:
            method_searching = "SDP_relaxation"
            if self._number_of_qubits<=20:
                method_searching='bruteforce'

            self._noise_strength_CTMP = fun_CTMP.find_optimal_noise_strength_CTMP(ctmp_rates_dictionary=CTMP_rates_dictionary,
                                                                                  number_of_qubits=self._number_of_qubits,
                                                                                  method_name=method_searching)

        self.__CTMP_generator_matrices = fun_CTMP.get_generators_dictionary_from_rates(
            rates_dictionary=CTMP_rates_dictionary)

    @property
    def CTMP_rates_dictionary(self) -> Dict[Tuple[int], Dict[str, float]]:
        return self._CTMP_rates_dictionary

    @CTMP_rates_dictionary.setter
    def CTMP_rates_dictionary(self, CTMP_rates_dictionary: Dict[Tuple[int], Dict[str, float]]) -> None:
        self._CTMP_rates_dictionary = CTMP_rates_dictionary


    def estimate_corrected_energy_CTMP_global(self,
                                              experiment_key=str,
                                              sampling_method='conditional',
                                              number_of_samples: Optional[int] = None,
                                              target_estimation_accuracy:Optional[float]=0.01,
                                              print_progress_bar=True
                                              ):
        """

        :param experiment_key:
        :type experiment_key:
        :param sampling_method:
        :type sampling_method:
        :param number_of_samples:
        :type number_of_samples:
        :param print_progress_bar:
        :type print_progress_bar:
        :return:
        :rtype:
        """

        # Choose sampling method.
        if sampling_method.upper() == 'CONDITIONAL':
            # If noise strength was not provided, calculate it here
            if self._noise_strength_CTMP is None:
                qprint("Noise strength not provided! Calculating now using SDP relaxation!")
                self._noise_strength_CTMP = fun_CTMP.find_optimal_noise_strength_CTMP(
                    ctmp_rates_dictionary=self._CTMP_rates_dictionary,
                    number_of_qubits=self._number_of_qubits,
                    method_name='SDP_relaxation',
                    method_kwargs={'hierarchy_level': 1})
                qprint('DONE!')
            noise_strength_CTMP = self._noise_strength_CTMP

        elif sampling_method.upper() == 'SIMPLE':
            noise_strength_CTMP = fun_CTMP.get_CTMP_error_rates_sum(
                rates_dictionary=self._CTMP_rates_dictionary)
            simple_noise_matrices_flat, probability_over_matrices_flat = CTMP_simp._get_inputs_to_sampling_algorithm(
                rates_dictionary=self._CTMP_rates_dictionary)

        else:
            raise ValueError(f"Sampling method '{sampling_method}' not recognized!")

        # theoretical_overhead = fun_CTMP.get_theoretical_overhead_CTMP(noise_strength_CTMP=noise_strength_CTMP)



        if number_of_samples is None:
            number_of_samples = fun_CTMP.get_theoretical_number_of_samples_CTMP(noise_strength_CTMP=noise_strength_CTMP,
                                                                                target_estimation_accuracy=target_estimation_accuracy)

            if number_of_samples > self._default_max_number_of_samples:
                number_of_samples = self._default_max_number_of_samples

        normalization_quasiprobability = np.exp(2 * noise_strength_CTMP)

        experimental_results = self._results_dictionary[experiment_key]
        hamiltonian_weights = self._hamiltonian_dictionaries[experiment_key]

        experimental_outcomes = list(experimental_results.keys())
        probability_distribution_over_outcomes = np.array(list(experimental_results.values()),
                                                          dtype=float)
        probability_distribution_over_outcomes /= np.sum(probability_distribution_over_outcomes)
        samples_uniform_outcomes = np.random.multinomial(n=number_of_samples,
                                                         pvals=probability_distribution_over_outcomes)

        outcomes_range = range(len(samples_uniform_outcomes))

        if print_progress_bar:
            outcomes_range = tqdm(outcomes_range)

        counts_dictionary = defaultdict(float)
        for index_outcome in outcomes_range:
            number_of_samples_from_outcome = samples_uniform_outcomes[index_outcome]
            bitstring_now = experimental_outcomes[index_outcome]
            matrix_powers_samples = np.random.poisson(lam=noise_strength_CTMP,
                                                      size=number_of_samples_from_outcome)

            for matrix_power in matrix_powers_samples:
                if (matrix_power % 2) == 0:
                    sign = 1.0
                else:
                    sign = -1.0

                if sampling_method.upper() == 'CONDITIONAL':
                    output_state = CTMP_cond.sample_from_power_of_B_matrix(
                        matrix_power=matrix_power,
                        input_state=bitstring_now,
                        generators_dictionary=self.__CTMP_generator_matrices,
                        noise_strength_CTMP=noise_strength_CTMP
                    )


                elif sampling_method.upper() == 'SIMPLE':
                    output_state = CTMP_simp.sample_from_power_of_B_matrix_simple(
                        global_input_state=bitstring_now,
                        matrix_power=matrix_power,
                        probability_over_matrices_list=probability_over_matrices_flat,
                        local_noise_matrices_list=simple_noise_matrices_flat)

                counts_dictionary[output_state] += sign

        energy_estimated = fda.estimate_energy_from_counts_dictionary(
            counts_dictionary=dict(counts_dictionary),
            weights_dictionary=hamiltonian_weights,
            normalize=False) * normalization_quasiprobability / number_of_samples

        self._energies_dictionary_corrected[experiment_key] = energy_estimated

        return energy_estimated



    def estimate_corrected_energy_CTMP_global_separable(self,
                                                        experiment_key: str,
                                                        clusters_list: List[Tuple[int]],
                                                        sampling_method='conditional',
                                                        number_of_samples: Optional[int] = None,
                                                        target_estimation_accuracy:Optional[float]=0.01,
                                                        print_progress_bar=True
                                                        ):
        """

        :param experiment_key:
        :type experiment_key:
        :param clusters_list:
        :type clusters_list:
        :param sampling_method:
        :type sampling_method:
        :param number_of_samples:
        :type number_of_samples:
        :param print_progress_bar:
        :type print_progress_bar:
        :return:
        :rtype:
        """

        # raise ValueError(f"Local sampling method not yet implemented")

        experimental_results = self._results_dictionary[experiment_key]
        hamiltonian_weights = self._hamiltonian_dictionaries[experiment_key]

        local_rates_dictionaries = self.__convert_to_local_noise_rates_dictionary(
            clusters_list=clusters_list)

        if sampling_method.upper() == 'SIMPLE':
            sampling_algorithm_inputs = {}
            for cluster, local_rates_dictionary in local_rates_dictionaries.items():
                sampling_algorithm_inputs[cluster] = CTMP_simp._get_inputs_to_sampling_algorithm(
                    rates_dictionary=local_rates_dictionary)

            noise_strength_method = 'SUM_OF_RATES'


        elif sampling_method.upper() == 'CONDITIONAL':

            local_generators_dictionary = {}
            for cluster, local_rates_dictionary in local_rates_dictionaries.items():
                local_generators_dictionary[cluster] = fun_CTMP.get_generators_dictionary_from_rates(
                    rates_dictionary=local_rates_dictionary)

            noise_strength_method = 'bruteforce'

        else:
            raise ValueError(
                f"Sampling method '{sampling_method}' unrecognized!\n Please use one of the following:"
                f" - 'SIMPLE'"
                f" - 'CONDITIONAL'")

        local_noise_strengths = self._calculate_noise_strengths_local(clusters_list=clusters_list,
                                                                      method=noise_strength_method)

        total_noise_strength = sum(list(local_noise_strengths.values()))
        normalization_quasiprobability = np.exp(2 * total_noise_strength)

        if number_of_samples is None:
            number_of_samples = fun_CTMP.get_theoretical_number_of_samples_CTMP(noise_strength_CTMP=total_noise_strength,
                                                                                target_estimation_accuracy=target_estimation_accuracy)

            if number_of_samples > self._default_max_number_of_samples:
                number_of_samples = self._default_max_number_of_samples







        experimental_outcomes = list(experimental_results.keys())
        probability_distribution_over_outcomes = np.array(list(experimental_results.values()),
                                                          dtype=float)
        probability_distribution_over_outcomes /= np.sum(probability_distribution_over_outcomes)
        samples_uniform_outcomes = np.random.multinomial(n=number_of_samples,
                                                         pvals=probability_distribution_over_outcomes)
        outcomes_range = range(len(samples_uniform_outcomes))

        if print_progress_bar:
            outcomes_range = tqdm(outcomes_range)

        counts_dictionary = defaultdict(float)
        for index_outcome in outcomes_range:
            number_of_samples_from_outcome = samples_uniform_outcomes[index_outcome]
            bitstring_now = experimental_outcomes[index_outcome]

            for _ in range(number_of_samples_from_outcome):
                global_output_state = [None for __ in range(self._number_of_qubits)]
                global_sign = 1
                for cluster in clusters_list:
                    # rates_dictionary_now = local_rates_dictionaries[cluster]
                    noise_strength_now = local_noise_strengths[cluster]

                    local_bitstring = [bitstring_now[qi] for qi in cluster]
                    local_matrix_power = np.random.poisson(lam=noise_strength_now,
                                                           size=1)[0]

                    if (local_matrix_power % 2) == 0:
                        global_sign *= 1.0
                    else:
                        global_sign *= -1.0

                    # print(rates_dictionary_now)

                    if sampling_method.upper() == 'CONDITIONAL':

                        generators_now = local_generators_dictionary[cluster]
                        local_output_state = CTMP_cond.sample_from_power_of_B_matrix(
                            matrix_power=local_matrix_power,
                            input_state=local_bitstring,
                            generators_dictionary=generators_now,
                            noise_strength_CTMP=noise_strength_now
                        )


                    elif sampling_method.upper() == 'SIMPLE':
                        local_output_state = CTMP_simp.sample_from_power_of_B_matrix_simple(
                            global_input_state=bitstring_now,
                            matrix_power=local_matrix_power,
                            probability_over_matrices_list=sampling_algorithm_inputs[cluster][1],
                            local_noise_matrices_list=sampling_algorithm_inputs[cluster][0])

                    for qubit_index in range(len(cluster)):
                        global_output_state[cluster[qubit_index]] = local_output_state[qubit_index]

                global_output_state = ''.join(global_output_state)
                counts_dictionary[global_output_state] += global_sign

        energy_estimated = fda.estimate_energy_from_counts_dictionary(
            counts_dictionary=dict(counts_dictionary),
            weights_dictionary=hamiltonian_weights,
            normalize=False) * normalization_quasiprobability / number_of_samples

        self._energies_dictionary_corrected[experiment_key] = energy_estimated

        return energy_estimated




"""
@authors: Filip Maciejewski, Oskar Słowik, Tomek Rybotycki
@contact: filip.b.maciejewski@gmail.com
"""
"""
@authors: Filip Maciejewski, Oskar Słowik, Tomek Rybotycki
@contact: filip.b.maciejewski@gmail.com
"""
# TODO OS: finish
import abc

class HeuristicClusteringAlgorithmBase(abc.ABC):

    @property
    @abc.abstractmethod
    def clusterize(self, **kwargs) -> Tuple[list, number]:
        # dictionary of experimental results
        raise NotImplementedError



#===================================================
# Constants
#===================================================
import numpy as np
#Why 9 is the number as of now? 
SIGNIFICANT_DIGITS = 9
SIGNIFICANT_DIGITS_POVM = 7

EXPERIMENT_TYPE_SYMBLOS = {"qdot": 6, "ddot": 2, "coh": 2}
CIRCUIT_DATA_TYPE = np.dtype(np.uint8) 


#For Circuits total count estimation:
CIRC_EFFECT_DISTANCE =0.1
CIRC_PROBABILITY_OF_FAIL = 0.5


"""
@author: Jan Tuziemski, Filip Maciejewski, Joanna Majsak, Oskar Słowik, Marcin Kotowski, Katarzyna Kowalczyk-Murynka, Paweł Przewłocki, Piotr Podziemski, Michał Oszmaniec
@contact: filip.b.maciejewski@gmail.com
"""

from typing import Optional
import time

import numpy as np
from qrem.functions_qrem import functions_data_analysis as fda

def __get_mapping_2SAT(clause_now,
                       j):
    variables = clause_now[1]
    signs = clause_now[0]
    if j not in variables:
        return 0
    else:
        location = int(np.where(np.array(variables) == j)[0])
        return signs[location]


def __get_local_field_1q(qubit_index,
                         clauses_list):
    variable_index = qubit_index

    local_field = -sum([__get_mapping_2SAT(ck, variable_index) for ck in clauses_list])

    return local_field

def __get_local_field_2q(qubits_indices,
                         clauses_list):
    variable_index0, variable_index1 = qubits_indices[0], qubits_indices[1]

    local_field = sum(
        [__get_mapping_2SAT(ck, variable_index0) * __get_mapping_2SAT(ck, variable_index1) for
         ck in clauses_list])

    return local_field


def get_random_2SAT_clauses(number_of_qubits,
                            clause_density):

    number_of_variables = number_of_qubits
    number_of_clauses = int(np.floor(clause_density * number_of_variables))

    clauses_list = []
    while len(clauses_list) < number_of_clauses:
        var_i = np.random.randint(0, number_of_qubits)
        var_j = np.random.randint(0, number_of_qubits)
        while var_i == var_j:
            var_j = np.random.randint(0, number_of_qubits)
        sign_i, sign_j = np.random.choice([-1, 1]), np.random.choice([-1, 1])

        clause_now = ((int(sign_i), int(sign_j)), tuple(sorted((int(var_i), int(var_j)))))
        if clause_now not in clauses_list:
            clauses_list.append(clause_now)


    return clauses_list


def get_weights_dictionary_from_weights_matrix(weights_matrix):

    weights_dictionary = {}


    number_of_qubits = weights_matrix.shape[0]

    for i in range(number_of_qubits):
        for j in range(i, number_of_qubits):
            if weights_matrix[i, j] != 0:
                if i == j:
                    weights_dictionary[(i,)] = weights_matrix[i, i]
                elif j > i:
                    weights_dictionary[(i, j)] = weights_matrix[i, j]

    return weights_dictionary

def get_2SAT_hamiltonian_from_clauses(clauses_list,
                                      number_of_qubits):


    weights_matrix = np.zeros((number_of_qubits, number_of_qubits))

    for i in range(number_of_qubits):
        for j in range(i, number_of_qubits):
            if i == j:
                weights_matrix[i,i]+=__get_local_field_1q(qubit_index=i,
                                                          clauses_list=clauses_list)


            else:
                weights_matrix[i,j]+= __get_local_field_2q(qubits_indices=[i, j],
                                                           clauses_list=clauses_list)

    weights_dictionary = get_weights_dictionary_from_weights_matrix(weights_matrix=weights_matrix)

    return weights_matrix, weights_dictionary





def generate_random_2SAT_hamiltonian(number_of_qubits:int,
                                     clause_density:float):


    random_clauses_list = get_random_2SAT_clauses(number_of_qubits=number_of_qubits,
                                                  clause_density=clause_density)


    weights_matrix, weights_dictionary = get_2SAT_hamiltonian_from_clauses(clauses_list=random_clauses_list,
                                                                           number_of_qubits=number_of_qubits)


    dictionary_to_return = {'clauses_list':random_clauses_list,
                            'number_of_qubits':number_of_qubits,
                            'weights_matrix':weights_matrix,
                            'weights_dictionary':weights_dictionary,
                            'clause_density':clause_density,
                            'name_of_hamiltonian':'2SAT'}

    return dictionary_to_return


def __change_representation_to_pysat(clause):
    return [clause[0][0]*(clause[1][0]+1),clause[0][1]*(clause[1][1]+1)]

def __get_state_from_other_format(solution):
    state = []
    for k in range(len(solution)):
        if np.sign(solution[k])==1:
            state.append('0')
        else:
            state.append('1')

    return state

def get_energy_from_SAT_solution(SATS, number_of_clauses):

    energy= 4*(-SATS+number_of_clauses)-number_of_clauses


    return energy

def __solve_2SAT_Hamiltonian_pysat(
        # weights_matrix,
                                   clauses_list,
                                   weights_dictionary,
                                   number_of_qubits,
                                   clause_density,
                                   repetitions = 1):

    from pysat.formula import WCNF
    from pysat.examples.rc2 import RC2Stratified

    new_rep = [__change_representation_to_pysat(cl) for cl in clauses_list]
    wcnf_now = WCNF()
    for cl in new_rep:
        wcnf_now.append(cl, weight=1)


    states_sol_repets, costs_sol_repets = [], []
    for heh in range(repetitions):
        solver_rc2 = RC2Stratified(wcnf_now,
                                   solver='g4')
        solution_this = solver_rc2.compute()
        sol_state_now = list(solver_rc2.model)
        sol_cost_now = solver_rc2.cost

        states_sol_repets.append(sol_state_now)
        costs_sol_repets.append(sol_cost_now)

    argmin_sol = np.argmin(costs_sol_repets)
    sol_cost_now = costs_sol_repets[argmin_sol]
    sol_state_now = states_sol_repets[argmin_sol]
    sat2_solution_rc2 = number_of_qubits * clause_density - sol_cost_now

    binary_state_rc2 = __get_state_from_other_format(sol_state_now)
    preferred_ground_state = ''.join([ex for ex in binary_state_rc2])
    # hamiltonian_elements = np.count_nonzero(weights_matrix)
    #
    # qubits, H1q, pairs, pairs_weights = fp.get_qubits_from_weights_matrix(weights_matrix)
    number_of_clauses = len(clauses_list)

    maxsat_solution, lowest_binary_state = sol_cost_now, preferred_ground_state

    # minimal_energy = get_energy_from_sat(maxsat_solution, number_of_clauses)
    min_energy = get_energy_from_SAT_solution(sat2_solution_rc2, number_of_clauses)

    list_state_binary = list(preferred_ground_state)
    ground_state_energy = fda.get_energy_from_bitstring_diagonal_local(bitstring=list_state_binary,
                                                                       weights_dictionary=weights_dictionary)\
        # .get_energy_from_string(list_state_binary, weights_dictionary)

    if abs(ground_state_energy - min_energy) >= 10 ** (-10):
        raise ValueError('something is wrong with ground state')

    dictionary_to_return = {
        # 'single_qubits': qubits,
        # 'pairs_of_qubits': pairs,
        # 'weighted_1q': H1q,
        # 'weighted_2q': pairs_weights,
        # 'weights_matrix': Wij,
        # 'weights_dictionary': weights_dict,
        # 'number_of_hamiltonian_elements': hamiltonian_elements,
        'ground_state': list_state_binary,
        'ground_state_energy': ground_state_energy,
        # 'clauses': clauses_list,
        'maxsat_solution': maxsat_solution
    }

    return dictionary_to_return





def solve_2SAT_hamiltonian(hamiltonian_data,
                           solving_method = 'pysat'
                           ):


    clauses_list = hamiltonian_data['clauses_list']
    weights_dictionary = hamiltonian_data['weights_dictionary']
    number_of_qubits = hamiltonian_data['number_of_qubits']
    clause_density = hamiltonian_data['clause_density']

    if solving_method.lower() in ['pysat']:
        solution_data = __solve_2SAT_Hamiltonian_pysat(clauses_list=clauses_list,
                                       weights_dictionary=weights_dictionary,
                                       number_of_qubits=number_of_qubits,
                                       clause_density=clause_density)

    return {**hamiltonian_data, **solution_data}


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: Jan Tuziemski, Filip Maciejewski, Joanna Majsak, Oskar Słowik, Marcin Kotowski, Katarzyna Kowalczyk-Murynka, Paweł Przewłocki, Piotr Podziemski, Michał Oszmaniec
email: filip.b.maciejewski@gmail.com

References:
[1] Filip B. Maciejewski, Zoltán Zimborás, Michał Oszmaniec, "Mitigation of readout noise in near-term quantum devices
by classical post-processing based on detector tomography", Quantum 4, 257 (2020).

[2] Zbigniew Puchała, Łukasz Pawela, Aleksandra Krawiec, Ryszard Kukulski, "Strategies for optimal single-shot
discrimination of quantum measurements", Phys. Rev. arrray_to_print 98, 042103 (2018), https://arxiv.org/abs/1804.05856

[3] T. Weissman, E. Ordentlich, G. Seroussi, S. Verdul, and M. J. Weinberger, Technical Report HPL-2003-97R1,
Hewlett-Packard Labs (2003).

[4] John A. Smolin, Jay M. Gambetta, Graeme Smith, "Maximum Likelihood, Minimum Effort", Phys. Rev. Lett. 108, 070502
(2012), https://arxiv.org/abs/1106.5458
"""

import numpy as np
from qrem.functions_qrem import povmtools

from typing import List, Optional
from qiskit.result import Result
from qrem.common.providers.ibmutils.data_converters import get_frequencies_array_from_results
from qrem.common import probabilty

class QDTErrorMitigator:
    """
        This class is used to mitigate errors in qiskit jobs_list using data from QDT.
    """

    def __init__(self):
        """
        Description:
            Constructor of the class. If povm is None, then it should be prepared before use (using prepare_mitigator)!
        """
        self.povm = None
        self.transition_matrix = None
        self.correction_matrix = None
        self.distances_from_closest_probability_vector = []
        self.qiskit_register_convention = False

        # if povm is not None:
        #     self.prepare_mitigator(povm)

    @staticmethod
    def correct_probability_vector(probability_vector:np.ndarray,
                                   correction_matrix:np.ndarray,
                                   ensure_physicality:Optional[bool]=True):

        corrected_statistics = correction_matrix.dot(probability_vector)
        number_of_povm_outcomes = len(probability_vector)


        if ensure_physicality:
            if not probabilty.is_valid_probability_vector(list(corrected_statistics[:, 0])):
                closest_physical_statistics = np.array(
                    povmtools.find_closest_prob_vector_l2(corrected_statistics)).reshape(
                    number_of_povm_outcomes, 1)
                corrected_statistics = closest_physical_statistics

        return corrected_statistics




    def prepare_mitigator(self, povm: List[np.ndarray]) -> None:
        """
        Description:
            This method_name is main method_name of the class. It is used to prepare the mitigation matrix from POVM. Other
            methods allow to calculate errors.

        Parameters:
            :param povm: POVM describing the detector used in measurements of which statistics are meant to be
             corrected.

        Returns:
            -

        """
        self.povm = povm
        self.__construct_transition_matrix()
        self.__construct_correction_matrix()

    def __construct_transition_matrix(self) -> None:
        """
        Description:
            Given classical description of the detector (i_index.e., matrix representation of POVM's elements), get the
            classical part of the noise from diagonal part of every effect. The classical part of the noise is
            represented by left-stochastic matrix. ASSUMING that ideal measurement is the von Neumann measurement in
            computational basis.
            See Ref. [1] for details. In Ref. [1] this matrix is denoted as \Lambda.

        Parameters:
            -

        Returns:
            arrray_to_print matrix representing classical part of the noise.
        """

        number_of_povm_outcomes = len(self.povm)
        dimension = self.povm[0].shape[0]

        self.transition_matrix = np.zeros((number_of_povm_outcomes, number_of_povm_outcomes), dtype=float)

        for k in range(number_of_povm_outcomes):
            current_povm_effect = self.povm[k]

            # Get diagonal part of the effect. Here we remove eventual 0 imaginary part to avoid format conflicts
            # (diagonal elements of Hermitian matrices are real).
            vec_p = np.array([np.real(current_povm_effect[i, i]) for i in range(dimension)])

            # Add vector to transition matrix.
            self.transition_matrix[k, :] = vec_p[:]

    def __construct_correction_matrix(self) -> None:
        """
        Description:
            Given classical description of the detector (i_index.e., matrix representation of POVM's elements), get the
            correction matrix based on classical part of the noise. ASSUMING that ideal measurement is the von Neumann
            measurement in computational basis. See Ref. [1] for details. In Ref. [1] this matrix is denoted as
            \Lambda^{-1}.

        Parameters:
            -

        Returns:
            :correction_matrix: numpy array representing correction matrix. It is the inverse of transition_matrix
            returned by function construct_transition_matrix
        """

        try:
            self.correction_matrix = np.linalg.inv(self.transition_matrix)
        except np.linalg.LinAlgError:
            print('Noise matrix is not invertible. Returning identity. Got:')
            print(self.transition_matrix)
            self.correction_matrix = np.eye(np.shape(self.transition_matrix[0]))

    def apply_correction_to_statistics(self, statistics_array: np.ndarray):
        """
        Description:
            Given correction matrix and vector of relative frequencies, correct the statistics via multiplication by
            correction_matrix.

            In case of obtaining quasiprobability vector after such correction, it is possible to find closest physical
            one (in Euclidean norm). Total Variation distance between quasi probability vector and closest physical one
            is the upper bound for the correction error resulting from this unphysicality. See Ref. [1] for details. In
            Ref.[1], such distance is denoted as \alpha.

        Parameters:
            :param statistics_array: Statistics for which correction should be performed.

        Returns:
            Corrected statistics.
        """

        number_of_povm_outcomes = self.correction_matrix[0].shape[0]

        number_of_qubits = int(np.log2(number_of_povm_outcomes))

        corrected_frequencies = []
        self.distances_from_closest_probability_vector = []

        for statistics in statistics_array:
            # make sure statistics have proper format
            statistics = np.array(statistics).reshape(number_of_povm_outcomes, 1)

            # Check if given statistics are normalized.
            norm = sum(statistics)
            if abs(norm - 1) >= 10 ** (-9):
                print('Warning: Frequencies are not normalized. We normalize them. They looked like this:')
                print(statistics)
                statistics = statistics / norm

            if self.qiskit_register_convention:
                # reverse statistics for time of correction
                statistics = povmtools.reorder_probabilities(statistics, range(number_of_qubits)[::-1])

                # make sure statistics have proper format
                statistics = np.array(statistics).reshape(number_of_povm_outcomes, 1)

            # corrected statistics by multiplication via correction matrix
            corrected_statistics = self.correct_probability_vector(statistics,
                                                                   self.correction_matrix,
                                                                   ensure_physicality=True)
            if self.qiskit_register_convention:
                # go back to standard convention
                corrected_statistics = povmtools.reorder_probabilities(corrected_statistics,
                                                                       range(number_of_qubits)[::-1])

                # make sure statistics have proper format
                corrected_statistics = np.array(corrected_statistics).reshape(number_of_povm_outcomes, 1)


        return corrected_frequencies

    def apply_correction_to_qiskit_job(self, results: Result) -> List[np.ndarray]:
        """
        Description:
            Given correction matrix and vector of relative frequencies, correct the statistics via multiplication by
            correction_matrix.

            In case of obtaining quasiprobability vector after such correction, it is possible to find closest physical
            one (in Euclidean norm). Total Variation distance between quasi probability vector and closest physical one
            is the upper bound for the correction error resulting from this unphysicality. See Ref. [1] for details. In
            Ref.[1], such distance is denoted as \alpha.

        Parameters:
            :param results: Qiskit job results for which statistics should be corrected.

        Returns:
            Corrected statistics.
        """

        # Create new object to avoid conflicts
        statistics_array = get_frequencies_array_from_results([results])
        return self.apply_correction_to_statistics(statistics_array)




import os
import time
from typing import Dict, Optional

import numpy as np
from tqdm import tqdm
import qiskit
from qiskit import Aer, IBMQ
from qiskit.providers.ibmq.job.ibmqjob import JobStatus

from qrem.functions_qrem.povmtools import reorder_probabilities
from qrem.functions_qrem import ancillary_functions as anf
from qrem.functions_qrem import povmtools

from qrem.common import io
from qrem.common.printer import qprint

#ORGANIZATION - poorly documented code (MO)

def get_qiskit_provider(backend_name: str,
                        provider_data: Dict[str, str]=None):
    IBMQ.load_account()
    if backend_name.lower() in ['ionq.qpu', 'ionq.simulator']:
        from azure.quantum.qiskit import AzureQuantumProvider

        if provider_data is None:
            AZURE_RESOURCE_ID = os.environ['AZURE_RESOURCE_ID_IONQ']
            AZURE_LOCATION = os.environ['AZURE_LOCATION_IONQ']
        else:
            AZURE_RESOURCE_ID = provider_data['AZURE_RESOURCE_ID_IONQ']
            AZURE_LOCATION = provider_data['AZURE_LOCATION_IONQ']

        provider = AzureQuantumProvider(
            resource_id=AZURE_RESOURCE_ID,
            location=AZURE_LOCATION)

    else:
        if provider_data is None:
            IBMQ_hub = os.environ['IBMQ_HUB']
            IBMQ_group = os.environ['IBMQ_GROUP']
            IBMQ_project = os.environ['IBMQ_PROJECT']
        else:
            IBMQ_hub = provider_data['IBMQ_HUB']
            IBMQ_group = provider_data['IBMQ_GROUP']
            IBMQ_project = provider_data['IBMQ_PROJECT']

        provider = IBMQ.get_provider(hub=IBMQ_hub,
                                     group=IBMQ_group,
                                     project=IBMQ_project)

    return provider

def add_gate_to_circuit(circuit,
                        qreg,
                        q_index,
                        unitary):
    # TODO: Check whether this is needed - at some point I remember there were troubles with parametrizing those two unitaries
    if unitary[0, 0] == 1 and unitary[1, 1] == 1:
        pass
    elif unitary[0, 1] == 1 and unitary[1, 0] == 1:
        circuit.X(qreg[q_index])

    else:
        # get angles for single-qubit state change unitary
        current_angles = povmtools.get_su2_parametrizing_angles(unitary)

        # implement unitary
        circuit.u3(current_angles[0],
                   current_angles[1],
                   current_angles[2],
                   qreg[q_index])
    return circuit, qreg


#ORGANIZE: copied to experiment_results. Delete when refactoring complete
def get_frequencies_from_counts(counts_dict,
                                crs=None,
                                classical_register=None,
                                shots_number=None,
                                reorder_bits=True):
    if crs is None:
        crs = len(list(list(counts_dict.keys())[0]))

    d = 2 ** crs

    if classical_register is None:
        classical_register = ["{0:b}".format(i).zfill(crs) for i in range(d)]

    normal_order = []

    for j in range(d):
        if classical_register[j] in counts_dict.keys():
            counts_now = counts_dict[classical_register[j]]
            normal_order.append(counts_now)

        else:
            normal_order.append(0)
    if reorder_bits:
        frequencies = reorder_probabilities(normal_order, range(crs)[::-1])
    else:
        frequencies = normal_order

    if shots_number is None:
        frequencies = frequencies / np.sum(frequencies)
    else:
        frequencies = frequencies / shots_number

    return frequencies


def download_multiple_jobs(backend_name,
                           job_IDs_list,
                           provider_data=None):
    IBMQ.load_account()
    if backend_name in ['qasm_simulator', 'statevector_simulator', 'unitary_simulator']:
        raise ValueError('Local simulators do not store jobs online.')

    provider = get_qiskit_provider(backend_name=backend_name,
                                   provider_data=provider_data)

    backend = provider.get_backend(backend_name)

    all_jobs = []
    for job_ID in tqdm(job_IDs_list):
        # qprint('Getting job with ID:', job_ID)
        job = backend.retrieve_job(job_ID)
        # qprint('Got it!')
        all_jobs.append(job)
    return all_jobs


#ORGANIZE: copied to experiment_results, but possibly not needed.
def add_counts_dicts(all_counts, modulo, dimension):
    frequencies = [np.zeros(dimension) for i in range(modulo)]

    for counts_index in tqdm(range(len(all_counts))):
        true_index = counts_index % modulo

        freqs_now = povmtools.counts_dict_to_frequencies_vector(all_counts[counts_index], True)
        frequencies[true_index][:] += freqs_now[:]

        # print(freqs_now)
    for i in range(modulo):
        frequencies[i] *= 1 / np.sum(frequencies[i])

    return frequencies

#MOcomm is it just for qiskit or more general

def run_batches(batches,
                backend_name,
                shots,
                saving_IDs_dictionary=dict(saving=False, directory=None, file_name=None, dictionary_to_save={}),
                provider_data:Optional[Dict[str,str]] = None, verbose_log = True):
    """
    Description:
        This function takes a list of circuit batches and sends them for execution on the specified backend.
        User can choose to save the resulting jobs' IDs, using parameters in saving_IDs_dictionary
    Parameters:
        :param batches: (list(list(qiskit.circuit.quantumcircuit.QuantumCircuit))): list of batches, where each batch is
        a list of qiskit circuits and each batch has length <= maximum number of circuits in job
        :param backend_name: (str): if value is of the names of a qiskit simulator: ('qasm_simulator',
        'statevector_simulator', 'unitary_simulator'), provider is chosen to be Aer; in other cases provider is taken
         from 'provider_data'
        #TODO: consider cases where other backend names are provided (e.g. aer_simulator)
        :param shots: number of repetitions of each circuit from batches
        :param saving_IDs_dictionary:
        :provider_data: Dict[str]: contains access data required by providers, like user ID etc. If None, as is default,
        this data is obtained from environmental variables by function get_qiskit_provider (see function for details)
    Returns:
        jobs: list of qiskit jobs (results of qiskit.execute())
    Notes:

    """
    saving = saving_IDs_dictionary['saving']

    qprint('\nSending jobs to execution on: ', backend_name + '.')
    qprint('Number of shots: ', str(shots) + ' .')
    qprint('Target number of jobs: ', str(len(batches)) + ' .')
    print()

    iterations_done = 0
    wait_time_in_minutes = 10
    jobs = []

    qiskit_simulators = ['qasm_simulator', 'statevector_simulator', 'unitary_simulator']
    braket_simulators = ['BraketLocalBackend']
    braket_quantum_devices = ['Aspen-M-1', 'Aspen-M-2']

    # set provider and backend according to passed argument 'backend_name':
    if backend_name in qiskit_simulators:
        backend = Aer.get_backend(backend_name)
    else:
        provider = get_qiskit_provider(backend_name=backend_name,
                                       provider_data=provider_data)
        backend = provider.get_backend(backend_name)

    # Run the circuits on backend:
    while iterations_done < len(batches):
        qprint('job number:', str(iterations_done))
        circuits = batches[iterations_done]
        try:
            time.sleep(2)
            qprint("Sending quantum program to: ", backend_name + '.')
            # below we use pure qiskit:
            if backend_name not in braket_simulators and backend_name not in braket_quantum_devices:
                job = qiskit.execute(circuits, backend, shots=shots, max_credits=200)#, qobj_id=qobj_id)

            # below we use braket (note that we use 'job' here in the qiskit sense, in braket this is called 'task'):
            else:
                job = backend.run(circuits, shots=shots, disable_qubit_rewiring=True)
                # save the circuit names as metadata to use in get_counts_from_jobs and get_counts_from_result_object:
                job.metadata['circuit_names'] = [circuit.name for circuit in circuits]
                # alternatives: job._tasks[idx]._result.task_metadata.braketSchemaHeader.name; Tag property of task
            jobs.append(job)

            if saving and backend_name not in ['qasm_simulator', 'statevector_simulator', 'unitary_simulator']:
                job_ID = job.job_id()
                dict_to_save = saving_IDs_dictionary['dictionary_to_save']
                dict_to_save['job_ID'] = job_ID

                io.save(dictionary_to_save=dict_to_save,
                         directory=saving_IDs_dictionary['directory'],
                         custom_name=saving_IDs_dictionary['file_name'] + '_job%s' % iterations_done,
                         overwrie = False)
            while job.status() == JobStatus.INITIALIZING:
                print(job.status())
                time.sleep(2)
            qprint("Program sent for execution to: ", backend_name + '.')

        except BaseException as ex:
            print('There was an error in the circuit!. Error = {}'.format(ex))
            print(f'Waiting {wait_time_in_minutes} minute(s) before next try.')
            time.sleep(wait_time_in_minutes * 60)
            continue

        print()
        iterations_done += 1
    return jobs




#ORGANIZE - would be great to know what is the role of all these packages
#--------------------------------------------------
# standard and technical imports 
#--------------------------------------------------
import os
import warnings
import re
import itertools #ORGANIZE - would be good if we trie dto be effitienta and used fast structures for iterations. @Piotr: this should be fast
import multiprocessing
import datetime as dt 
from collections import defaultdict
from typing import List, Dict, Optional, Callable, Tuple
import pickle
 
#--------------------------------------------------
# scientific imports
#--------------------------------------------------
import cmath as c
import numpy as np
import pandas as pd
from colorama import Fore, Style

from qrem.common.printer import qprint
from qrem.common import utils
from qrem.common import convert
#--------------------------------------------------
# local imports
#--------------------------------------------------
#NONE

    





#MOVE_TO >> core.utils
# core.utils.DitstringConverter
# MO - check if usage makes sense? Can be iefficient
def get_qubits_keystring(list_of_qubits: List[int]) -> str:
    """ from subset of qubit indices get the string that labels this subset
        using convention 'q5q6q12...' etc.
    :param list_of_qubits: labels of qubits

    :return: string label for qubits

     NOTE: this function is "inverse operation" to get_qubit_indices_from_keystring.
    """

    if list_of_qubits is None:
        return None
    return 'q' + 'q'.join([str(s) for s in list_of_qubits])

#MOVE_TO >> core.utils core.numerics
def all_possible_bitstrings_of_length(number_of_bits: int,
                reversed: Optional[bool] = False):
    """Generate outcome bitstrings for n-qubits.

    Args:
        number_of_qubits (int): the number of qubits.

    Returns:
        list: arrray_to_print list of bitstrings ordered as follows:
        Example: n=2 returns ['00', '01', '10', '11'].
"""
    if (reversed == True):
        return [(bin(j)[2:].zfill(number_of_bits))[::-1] for j in list(range(2 ** number_of_bits))]
    else:
        return [(bin(j)[2:].zfill(number_of_bits)) for j in list(range(2 ** number_of_bits))]

#MOVE_TO >> core.utils core.numerics (PP) core.utils.DitConverter
#(PP): compare with povmtools version and usage across package
#(PP): change name to bitstring inside
def get_classical_register_bitstrings(qubit_indices: List[int],
                                      quantum_register_size: Optional[int] = None,
                                      rev: Optional[bool] = False):
    """
    Register of qubits of size quantum_register_size, with only bits corresponding to qubit_indices
    Gets list of bitstrings of length quantum_register_size when only bits in qubit_indicies can be 0 and 1, others have to be 0

    Qubits indices are always indexed from right to get proper output, use ref if you need a format with indexing from left.

    use rev when input indices were 


    varying

    :param qubit_indices:
    :param quantum_register_size:
    :param rev:
    :return:
    """

    # TODO FBM: refactor this function.

    # again assumes that qubit_indices contains unique values
    if quantum_register_size is None:
        quantum_register_size = len(qubit_indices)

    if quantum_register_size == 0:
        return ['']

    if quantum_register_size == 1:
        return ['0', '1']

    all_bitstrings= all_possible_bitstrings_of_length(quantum_register_size, rev)
    not_used = []

    for j in list(range(quantum_register_size)):
        if j not in qubit_indices:
            not_used.append(j)

    bad_names = []
    for bitstring in all_bitstrings: #0000111
        for k in (not_used):
            rev_name = bitstring[::-1] #1110000 reverses order of string - why?
            if rev_name[k] == '1':
                bad_names.append(bitstring)

    relevant_names = []
    for bitstring in all_bitstrings:
        if bitstring not in bad_names:
            relevant_names.append(bitstring)

    return relevant_names

#MOVE_TO >> core.utils core.numerics (PP) core.utils.DitConverter
def register_names_qubits(qubit_indices: List[int],
                          quantum_register_size: Optional[int] = None,
                          rev: Optional[bool] = False):
    # depreciated
    return get_classical_register_bitstrings(qubit_indices=qubit_indices,
                                             quantum_register_size=quantum_register_size,
                                             rev=rev)

#MOVE_TO >> core.utils? DELETE
def convert_qubits_string_to_tuple(qubits_string):
    return tuple(convert.get_qubit_indices_from_keystring(qubits_string))
    
#MOVE_TO >> core.constans core.backends
def get_historical_experiments_number_of_qubits(backend_name: str):
    if backend_name == 'ibmq_16_melbourne':
        number_of_qubits = 15
    elif backend_name == 'ASPEN-8':
        number_of_qubits = 23
    elif backend_name.upper() == 'ASPEN-9':
        number_of_qubits = 20
    else:
        raise ValueError('Wrong backend name')

    return number_of_qubits

     


# =======================================================================================
# NUTILS
# =======================================================================================




# =======================================================================================
# IO FILESYSTEM
# =======================================================================================

# #MOVE_TO >> core.io
# def get_local_storage_directory():
#     # please create the environment variable in your virtual environment!
#     # print(get_local_storage_directory())
#     # print(os.listdir(os.environ['LOCAL_STORAGE_LOCATION']))
#     # print(os.environ.keys())

#     return os.environ['LOCAL_STORAGE_LOCATION']


# #MOVE_TO >> core.io
# def open_file_pickle(file_path):
#     with open(file_path, 'rb') as filein:
#         data_object = pickle.load(filein)

#     return data_object



# #MOVE_TO >> core.io TO DELETE PP CAN BE A MARKER FOR UNUSED STUFF
# def save_results_pickle(dictionary_to_save: dict,
#                         directory: str,
#                         custom_name: Optional[str] = None,
#                         get_cwd: Optional[bool] = False):
    
#     from pathlib import Path
#     if directory :
#         dir = Path(directory) 
#         dir.mkdir(mode=0o777, exist_ok=True, parents= True) 
#     else:
#         dir = Path.home().joinpath("qrem_results")
#         dir.mkdir(mode=0o777, exist_ok=True) 
    

#     # if (directory != None):
#     #     #!?! add / at the end ...
#     #     fp0 = [s for s in directory]
#     #     if (fp0[len(fp0) - 1] != '/'):
#     #         fp0.append('/')
#     #     fp = ''.join(fp0)
#     # else:
#     #     fp = ''

#     # Time& date
#     if (get_cwd):
#         cwd = os.getcwd()
#     else:
#         cwd = ''
#     ct3 = gate_proper_date_string()
#     # original_umask = os.umask(0)
#     # os.umask(original_umask)

#     main_directory = cwd + fp

#     # permission_mode=int('0777',8)
#     # os.chmod(main_directory,permission_mode)
#     check_dir = os.path.exists(main_directory)

#     if (check_dir == False):
#         try:
#             os.makedirs(main_directory)

#         except(FileExistsError):
#             import shutil
#             try:
#                 shutil.rmtree(main_directory)
#                 os.makedirs(main_directory)
#             except(FileExistsError):
#                 os.makedirs(main_directory)

#         print( Fore.CYAN + Style.BRIGHT + 'Attention: ' + Style.RESET_ALL + 'Directory ' + '"' + main_directory + '"' + ' was created.')
#         # os.umask(oldmask)
#         # os.chmod(main_directory,permission_mode)
#     if custom_name is None:
#         file_path = str(main_directory + 'Results_Object' + '___' + str(ct3))
#     else:
#         file_path = str(main_directory + custom_name)

#     # os.chmod(main_directory)
#     dict_results = dictionary_to_save

#     add_end = ''

#     if (file_path[len(file_path) - 4:len(file_path)] != '.pkl'):
#         add_end = '.pkl'

#     with open(file_path + add_end, 'wb') as f:
#         pickle.dump(dict_results, f, pickle.HIGHEST_PROTOCOL)












# =======================================================================================
# CANDIDATES TO BE DELETED BELOW
# =======================================================================================

#MOVE_TO >> core.utils / duplicate
# #ORGANIZE - this function is used only once for some flder opeprations (MO)
# def get_drive_path():
#     try:
#         return os.environ['LOCAL_DRIVE_DIRECTORY']
#     except(KeyError):
#         return os.environ['LOCAL_STORAGE_LOCATION']


#MOVE_TO >>  DELETE
# def generate_classical_register_bistrings_product(clusters_list: List[Tuple[int]]):
#     clusters_lengths = [len(cl) for cl in clusters_list]
#     if np.max(clusters_lengths) > 15:
#         raise ValueError(f"Too big cluster with size {np.max(clusters_lengths)}")

#     # number_of_qubits = sum(clusters_lengths)

#     local_registers = {cluster: get_classical_register_bitstrings(qubit_indices=range(len(cluster)))
#                        for cluster in clusters_list}
#     # for cluster in clusters_list:
#     #     register_cluster =
#     #     local_registers.append([(cluster, x) for x in register_cluster])
#     #
#     # cartesian_product = itertools.product(local_registers)
#     #
#     # bitstrings_list = []
#     #
#     # for bitstrings_list_with_cluster in cartesian_product:
#     #     bitstring_global = ['0' for _ in range(number_of_qubits)]
#     #     print(bitstrings_list_with_cluster[0])
#     #     for cluster, bitstring_local in bitstrings_list_with_cluster[0]:
#     #         # print(cluster, bitstring_local)
#     #         for index in range(len(bitstring_local)):
#     #             bitstring_global[cluster[index]] = bitstring_local[index]
#     #
#     #     bitstrings_list.append(bitstring_global)
#     return local_registers



#(PP)ORGANIZE - this function is i) trivial and ii)  nerer used - a cleare candidate for deletion  
#def decompose_qubit_matrix_in_pauli_basis(matrix):
#    return {pauli_label: 1 / 2 * np.trace(matrix @ pauli_matrix)
#            for pauli_label, pauli_matrix in pauli_sigmas.items()}

# def get_bell_basis():
#     return list(___bell_states___.values())


#(PP)ORGANIZE - this class is not used anywhere - and hence it is a candidate for deletion (MO)
# (PP) Also, what class does in the function folder?

# class key_dependent_dict(defaultdict):
#     """
#     This is class used to construct dictionary which creates values of keys in situ, in case there
#     user refers to key that is not present.

#     COPYRIGHT NOTE
#     This code was taken from Reddit thread:
#     https://www.reddit.com/r/Python/comments/27crqg/making_defaultdict_create_defaults_that_are_a/

#     """

#     def __init__(self, f_of_x=None):
#         super().__init__(None)  # base class doesn't get a factory
#         self.f_of_x = f_of_x  # save f(x)

#     def __missing__(self, key):  # called when a default needed
#         ret = self.f_of_x(key)  # calculate default value
#         self[key] = ret  # and install it in the dict
#         return ret


#ORGANIZE: this function does not appear anywther - candidate for deletion

#def get_time_now():
#    ct0 = str(dt.datetime.today())     #MO: in this lane I replaced datetime.datetime.today() to dt.datetime.today() need to return to previous verion if does not work
#    ct1 = str.replace(ct0, ':', '_')
#    ct2 = str.replace(ct1, '.', '_')
#    ct3 = ct2[0:19]
#
#    return ct3



#ORGANIZE: this function does not appear anywther - candidate for deletion (MO)

#def wrapped_multiprocessing_function_depreciated(tuple_of_main_arguments: List[tuple],
#                                                 additional_kwargs: dict,
#                                                 function_to_multiprocess: Callable,
#                                                 number_of_threads: Optional[int] = None,
#                                                 # debug=False,
#                                                 printing=False):
#    """
#
#    :param tuple_of_main_arguments:
#
#    This is list of tuples that is divided into batches that are passed to different threads of
#    multiprocessing. This, therefore should be set of VARIABLE arguments that are passed to the
#    function of interest
#
#    :param additional_kwargs:
#    This is dictionary of arguments that are CONSTANT for all function evaluations.
#    The dictionary is COPIED for each thread and passed to the function.
#
#
#    :param function_to_multiprocess:
#    :param number_of_threads:
#    :return:
# """         if number_of_threads is None:
#         number_of_threads = int(
#             np.min([multiprocessing.cpu_count() - 1, len(tuple_of_main_arguments)]))

#     length = len(tuple_of_main_arguments)

#     # IF there is less arguments than threads, we reduce number of threads
#     if length < number_of_threads:
#         number_of_threads = length

#     division_cores = length // number_of_threads

#     all_indices, arguments = [], []

#     # if debug:
#     # print(length, division_cores)

#     for process_pool_index in range(number_of_threads):
#         if process_pool_index == number_of_threads - 1:
#             slice_now = slice(process_pool_index * division_cores,
#                               -1)
#         else:
#             slice_now = slice(process_pool_index * division_cores,
#                               (process_pool_index + 1) * division_cores)

#         sample_indices_now = tuple_of_main_arguments[slice_now]
#         arguments.append((sample_indices_now,
#                           additional_kwargs))

#     # if debug:
#     #     raise KeyError
#     if printing:
#         qprint(f'Running {number_of_threads} threads!', '', 'green')
#     pool = multiprocessing.Pool(number_of_threads)
#     results = pool.starmap_async(function_to_multiprocess,
#                                  arguments)
#     pool.close()
#     pool.join()

#     res_multiprocessing = results.get()

#     results_dictionary_all = {}

#     for dict_now in res_multiprocessing:
#         results_dictionary_all = {**results_dictionary_all, **dict_now}

#     # last_run_indices = list(range(int(division_cores * number_of_threads), length))
#     # last_mp = [([tuple_of_main_arguments[x]], additional_kwargs) for x in last_run_indices]

#     # if debug:
#     #     print('hey bro', last_run_indices, len(last_run_indices))

#     return results_dictionary_all

# """


#ORGANIZE - the function below is not used throughout the project - I commented it out (MO)

#def try_to_do_something_multiple_times(what_to_do: Callable,
#                                       print_string: str,
#                                       how_many_tries=1000):
    # """
    # Codes try to run "what_to_do" callable "how_many_tries" times,
    # waiting 10 seconds each time there is an error.

    # :param what_to_do:
    # :param print_string:
    # :return:
    # """

#    for error_counter in range(how_many_tries):
#        try:
#           return what_to_do()
#        except:
#            print(
#                f"WARNING: An error occurred for experiment: '{print_string}'.")
#            traceback.print_exc()
#
 #       print(f"Waiting {10}s.")
  #      time.sleep(10)
   #     print(f"Retrying experiment: '{print_string}'.")
#
#        if (error_counter + 1) % 25 == 0:
#            qprint(
#                f"\n\nAlready tried '{print_string}' {str(error_counter)} times! Something seems to be wrong...\n\n",
#                '',
#                'red')